/*
 *   stunnel       Universal SSL tunnel
 *   Copyright (c) 1998-2006 Michal Trojnara <Michal.Trojnara@mirt.net>
 *                 All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   In addition, as a special exception, Michal Trojnara gives
 *   permission to link the code of this program with the OpenSSL
 *   library (or with modified versions of OpenSSL that use the same
 *   license as OpenSSL), and distribute linked combinations including
 *   the two.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than OpenSSL.  If you modify
 *   this file, you may extend this exception to your version of the
 *   file, but you are not obligated to do so.  If you do not wish to
 *   do so, delete this exception statement from your version.
 */

/* Undefine if you have problems with make_sockets() */
#define INET_SOCKET_PAIR

#include "common.h"
#include "prototypes.h"

#ifdef SSPI
#ifdef _WIN32
    #include <windows.h>
    #include <winsock.h>
    #include <wincrypt.h>
    #include <wintrust.h>
    #include <schannel.h>
    #include <time.h>
    #define SECURITY_WIN32
    #include <security.h>
    #include <sspi.h>
    #include <WinCryptEx.h>
    #define NT4_DLL_NAME L"Security.dll"
#else
    #ifdef HSM_SERVER
    #undef USE_LIBWRAP
    #endif
    #include "CSP_WinDef.h"
    #include "../../../src/fenix/hsm_ca.h"
    #include "CSP_WinCrypt.h"
    #include "CSP_Sspi.h"
    #include "CSP_SChannel.h"
    #include "CpSSP.h"
    #ifdef IOS
    extern DWORD CSP_GetLastError(void);
    #define GetLastError() CSP_GetLastError()
    #endif

#endif //_WIN32
#endif //SSPI

#ifndef UNUSED
#define UNUSED(x) ((void)(x))
#endif

#ifndef SHUT_RD
#define SHUT_RD 0
#endif
#ifndef SHUT_WR
#define SHUT_WR 1
#endif
#ifndef SHUT_RDWR
#define SHUT_RDWR 2
#endif

/* TCP wrapper */
#ifdef USE_LIBWRAP
#include <tcpd.h>
int allow_severity=LOG_NOTICE;
int deny_severity=LOG_WARNING;
#endif

#if SSLEAY_VERSION_NUMBER >= 0x0922
static char *sid_ctx="stunnel SID";
    /* const allowed here */
#endif

#ifdef SSPI
#define LocalAlloc(a,b) malloc(b)
#define IS_SOCKET_ERROR(a) (a<0)
#ifndef _WIN32
    static int WSAGetLastError()
    {
	return errno;
    }
    #define MoveMemory memmove
#else
HMODULE g_hSecurity = NULL;
#endif
#define LocalFree free

DWORD   dwProtocol      = 0;

SecPkgContext_StreamSizes Sizes;
void print_state(CLI* c);
static void verify_remote_cert(CLI* c);
static BOOL ssl_verify_by_level(CLI* c, PCCERT_CONTEXT pCert);
#endif //SSPI

static void do_client(CLI *);
static void run_client(CLI *);
static void init_local(CLI *);
static void init_remote(CLI *);
static void init_ssl(CLI *);
static void transfer(CLI *);
static void parse_socket_error(CLI *, const char *);

#ifndef SSPI
static void print_cipher(CLI *);
#endif //SSPI
static void auth_libwrap(CLI *);
static void auth_user(CLI *);
static int connect_local(CLI *);
#ifndef USE_WIN32
static void make_sockets(CLI *, int [2]);
#endif
static int connect_remote(CLI *);
int connect_wait(CLI *);
static void reset(int, char *);

int max_clients;
#ifndef USE_WIN32
int max_fds;
#endif
void s_dump(int level, const char *pb, int cb);

/* //need if you want to debug protocol 
static void 
PrintHexDump(DWORD length, PCHAR buffer)
{
    DWORD i,count,ind;
    CHAR rgbDigits[]="0123456789abcdef";
    CHAR rgbLine[100];
    char cbLine;

    for(ind = 0; length; length -= count, buffer += count, ind += count) 
    {
        count = (length > 16) ? 16:length;

        sprintf(rgbLine, "%4.4x  ",ind);
        cbLine = 6;

        for(i=0;i<count;i++) 
        {
            rgbLine[cbLine++] = rgbDigits[(buffer[i] >> 4) & 0x0f];
            rgbLine[cbLine++] = rgbDigits[buffer[i] & 0x0f];
            if(i == 7) 
            {
                rgbLine[cbLine++] = ':';
            } 
            else 
            {
                rgbLine[cbLine++] = ' ';
            }
        }
        for(; i < 16; i++) 
        {
            rgbLine[cbLine++] = ' ';
            rgbLine[cbLine++] = ' ';
            rgbLine[cbLine++] = ' ';
        }

        rgbLine[cbLine++] = ' ';

        for(i = 0; i < count; i++) 
        {
            if(buffer[i] < 32 || buffer[i] > 126 || buffer[i] == '%') 
            {
                rgbLine[cbLine++] = '.';
            } 
            else 
            {
                rgbLine[cbLine++] = buffer[i];
            }
        }

        rgbLine[cbLine++] = 0;
	s_dump(LOG_NOTICE, rgbLine, cbLine-1);
    }
} // End of  PrintHexDump
*/
#ifdef UNIX
//if stunnel on unix socket is client side
typedef HRESULT (*WireSendFDnEUID_ptr) (int socket_desc);
static int use_passcred1(CLI *c) {
	  TSupModuleInstance handle = support_load_library_registry 
		(LIBDIR "/libcspr.so", SUP_LOAD_LIBRARY_DEFAULT | SUP_LOAD_LIBRARY_SEPARATE|SUP_LOAD_LIBRARY_LAZY);
	  WireSendFDnEUID_ptr WireSendFDnEUID_func = 
		(WireSendFDnEUID_ptr) support_load_library_getaddr( handle, "WireSendFDnEUID" );

	  if(!WireSendFDnEUID_func || WireSendFDnEUID_func(c->fd))
            return 1;
	  return 0;
}

typedef HRESULT (*WireRecvFDnEUID_ptr) (int socket_desc, uid_t *cl_uid, gid_t *cl_gid);
//if stunnel on unix socket is server side
static int impersonate(int  fd) {
    uid_t uid, gid;
    TSupModuleInstance handle = support_load_library_registry 
        (LIBDIR "/libcspr.so", SUP_LOAD_LIBRARY_DEFAULT | SUP_LOAD_LIBRARY_SEPARATE|SUP_LOAD_LIBRARY_LAZY);
    WireRecvFDnEUID_ptr WireRecvFDnEUID_func = 
        (WireRecvFDnEUID_ptr) support_load_library_getaddr( handle, "WireRecvFDnEUID" );

    if(!WireRecvFDnEUID_func || WireRecvFDnEUID_func(fd, &uid , &gid))
        return 1;
    return 0;
}

#endif //UNIX

/* Allocate local data structure for the new thread */
void *alloc_client_session(LOCAL_OPTIONS *opt, int rfd, int wfd) {
    CLI *c;

    c=calloc(1, sizeof(CLI));
    if(!c) {
        s_log(LOG_ERR, "Memory allocation failed");
        return NULL;
    }
    c->opt=opt;
    c->local_rfd.fd=rfd;
    c->local_wfd.fd=wfd;
    return c;
}

void *client(void *arg) {
    CLI *c=arg;
    s_log(LOG_DEBUG,"client start");

#ifdef DEBUG_STACK_SIZE
    stack_info(1); /* initialize */
#endif
    s_log(LOG_DEBUG, "%s started", c->opt->servname);
#ifndef USE_WIN32
    if(c->opt->option.remote && c->opt->option.program) {
            /* connect and exec options specified together */
            /* -> spawn a local program instead of stdio */
        while((c->local_rfd.fd=c->local_wfd.fd=connect_local(c))>=0) {
            run_client(c);
            sleep(1); /* FIXME: not a good idea in ucontext threading */
        }
    } else
#endif
    {
#if defined (SSPI) && !defined(USE_WIN32)	
    if(c->opt->local_sockaddr && options.option.for_hsm)
    {
        if (impersonate(c->local_rfd.fd))
        {
            s_log(LOG_ERR,"impersonate on unix socket failed");
            exit(1);
        }
        s_log(LOG_NOTICE,"impersonate OK");
    }
#endif //SSPI && !_WIN32    
        if(alloc_fd(c->local_rfd.fd))
            return NULL;
        if(c->local_wfd.fd!=c->local_rfd.fd)
            if(alloc_fd(c->local_wfd.fd))
                return NULL;
        run_client(c);
    }
    free(c);
#ifdef DEBUG_STACK_SIZE
    stack_info(0); /* display computed value */
#endif
#if defined(USE_WIN32) && !defined(_WIN32_WCE)
    _endthread();
#endif
#ifdef USE_UCONTEXT
    s_log(LOG_DEBUG, "Context %ld closed", ready_head->id);
    s_poll_wait(NULL, 0, 0); /* wait on poll() */
    s_log(LOG_ERR, "INTERNAL ERROR: failed to drop context");
#endif
    return NULL;
}

static void run_client(CLI *c) {
    int error;

    c->remote_fd.fd=-1;
    c->fd=-1;
#ifndef SSPI
    c->ssl=NULL;
#endif //SSPI
    c->sock_bytes=c->ssl_bytes=0;

    ios_run_watch_callback(1,0,0);
    error=setjmp(c->err);
    if(!error)
        do_client(c);
#ifdef USE_PAM
    if(c->pamh)
        pam_end(c->pamh,PAM_SUCCESS);
#endif //USE_PAM

    s_log(LOG_NOTICE,
        "Connection %s: %d bytes sent to SSL, %d bytes sent to socket",
         error ? "reset" : "closed", c->ssl_bytes, c->sock_bytes);

#ifdef SSPI // release all resorse 
	s_log(LOG_DEBUG,"free Buffers");
	free(c->ExDataBuffer);
	free(c->wIoBuffer);
	free(c->rIoBuffer);
	free(c->writeBuffer);
	s_log(LOG_DEBUG,"delete c->hContext");
	g_pSSPI->DeleteSecurityContext(&c->hContext);
	if(c->opt->option.client) {
	    s_log(LOG_DEBUG,"delete c->hClientCreds");
	    g_pSSPI->FreeCredentialsHandle(&c->hClientCreds);
	}
	if(c->opt->hProv) {
	    s_log(LOG_DEBUG,"release c->opt->hProv");
	    CryptReleaseContext(c->opt->hProv,0);
	}
        if(c->opt->option.client) {
	    if(c->hMyCertStore)
		CertCloseStore(c->hMyCertStore,CERT_CLOSE_STORE_FORCE_FLAG);
	    if(c->hSysCertStore)
		CertCloseStore(c->hSysCertStore,CERT_CLOSE_STORE_FORCE_FLAG);
        }
	s_log(LOG_NOTICE,"incomp_mess = %d, extra_data = %d",c->incomp_mess,c->extra_data);
#endif //SSPI

        /* Cleanup IDENT socket */
    if(c->fd>=0)
        closesocket(c->fd);

        /* Cleanup SSL */
#ifndef SSPI
    if(c->ssl) { /* SSL initialized */
        SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
        SSL_free(c->ssl);
        ERR_remove_state(0);
    }
#endif //SSPI

        /* Cleanup remote socket */
    if(c->remote_fd.fd>=0) { /* Remote socket initialized */
        if(error && c->remote_fd.is_socket)
            reset(c->remote_fd.fd, "linger (remote)");
        closesocket(c->remote_fd.fd);
    }

        /* Cleanup local socket */
    if(c->local_rfd.fd>=0) { /* Local socket initialized */
        if(c->local_rfd.fd==c->local_wfd.fd) {
            if(error && c->local_rfd.is_socket)
                reset(c->local_rfd.fd, "linger (local)");
            closesocket(c->local_rfd.fd);
        } else { /* STDIO */
            if(error && c->local_rfd.is_socket)
                reset(c->local_rfd.fd, "linger (local_rfd)");
            if(error && c->local_wfd.is_socket)
                reset(c->local_wfd.fd, "linger (local_wfd)");
       }
    }
    ios_run_watch_callback(-1,0,0);
#ifdef USE_FORK
#if defined  SSPI && defined HSM_SERVER
exit(0);
#endif
    if(!c->opt->option.remote) /* 'exec' specified */
        child_status(); /* null SIGCHLD handler was used */
#else
    enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
    s_log(LOG_DEBUG, "%s finished (%d left)", c->opt->servname,
        --num_clients);
    leave_critical_section(CRIT_CLIENTS);
#endif
}

#ifdef  USE_PAM
#ifdef SOLARIS
int st_pamconv_func(int num_msg, struct pam_message **msg,
		              struct pam_response **resp, void *appdata_ptr)
#else
int st_pamconv_func(int num_msg, const struct pam_message **msg,
		              struct pam_response **resp, void *appdata_ptr)
#endif
{
    return PAM_SUCCESS;
}
#endif //USE_PAM
static void do_client(CLI *c) {
    init_local(c);
#ifdef SSPI   //allocate resources 
    c->opt->hProv = 0;
    c->ExDataBuffer = calloc(2*BUFFSIZE+20,1);
    if (!c->ExDataBuffer){
	s_log(LOG_ERR,"calloc failed for c->ExDataBuffer");
	longjmp(c->err,1);
    }
    c->wIoBuffer = calloc(2*BUFFSIZE+100,1);
    if (!c->wIoBuffer){
	s_log(LOG_ERR,"calloc failed for c->wIoBuffer");
	longjmp(c->err,1);
    }
    c->rIoBuffer = calloc(2*BUFFSIZE+100,1);
    if (!c->rIoBuffer){
	s_log(LOG_ERR,"calloc failed for c->rIoBuffer");
	longjmp(c->err,1);
    }
    c->writeBuffer = calloc(2*BUFFSIZE+100,1);
    if (!c->writeBuffer){
	s_log(LOG_ERR,"calloc failed for c->writeBuffer");
	longjmp(c->err,1);
    }
    if(c->opt->option.client){
#ifdef HSM_SERVER
        //need open store on 999 even when optimization is on
        setegid(999); 
        seteuid(999); 
#endif
    c->hSysCertStore = CertOpenStore(CERT_STORE_PROV_SYSTEM,PKCS_7_ASN_ENCODING|X509_ASN_ENCODING,
		    0,CERT_SYSTEM_STORE_LOCAL_MACHINE,L"MY");
#ifdef UNICODE
    c->hMyCertStore = CertOpenSystemStore(0, L"MY");
#else
    c->hMyCertStore = CertOpenSystemStore(0, "MY");
#endif //UNICODE
    if(!c->hMyCertStore && !c->hSysCertStore){
        s_log(LOG_ERR,"Error 0x%x returned by CertOpenSystemStore\n", 
        GetLastError());
#ifdef HSM_SERVER
        seteuid (0); setegid (0);
#endif
        longjmp(c->err,1);
    }
#ifdef HSM_SERVER
        seteuid (0);
        setegid (0);
#endif
    }
#endif //SSPI
    if(!c->opt->option.client && !c->opt->protocol) {
        /* Server mode and no protocol negotiation needed */
        init_ssl(c);
        init_remote(c);
    } else {
        init_remote(c);
        negotiate(c);
        init_ssl(c);
    }
#ifdef SSPI
//get size of TLS record header and trailer.
    if(g_pSSPI->QueryContextAttributes(&c->hContext, SECPKG_ATTR_STREAM_SIZES, &Sizes) != SEC_E_OK)
    {
        s_log(LOG_ERR,"Couldn't get Sizes");
        longjmp(c->err,1);
    }
#endif //SSPI    
    transfer(c);
}

static void init_local(CLI *c) {
    SOCKADDR_UNION addr;
    socklen_t addrlen;

    addrlen=sizeof(SOCKADDR_UNION);
    if(getpeername(c->local_rfd.fd, &addr.sa, &addrlen)<0) {
        strcpy(c->accepting_address, "NOT A SOCKET");
        c->local_rfd.is_socket=0;
        c->local_wfd.is_socket=0; /* TODO: It's not always true */
#ifdef USE_WIN32
        if(get_last_socket_error()!=ENOTSOCK) {
#else
        if(c->opt->option.transparent || get_last_socket_error()!=ENOTSOCK) {
#endif
            sockerror("getpeerbyname");
            longjmp(c->err, 1);
        }
        /* Ignore ENOTSOCK error so 'local' doesn't have to be a socket */
    } else { /* success */
        /* copy addr to c->peer_addr */
        memcpy(&c->peer_addr.addr[0], &addr, sizeof(SOCKADDR_UNION));
        c->peer_addr.num=1;
        s_ntop(c->accepting_address, &c->peer_addr.addr[0]);
        c->local_rfd.is_socket=1;
        c->local_wfd.is_socket=1; /* TODO: It's not always true */
        /* It's a socket: lets setup options */
        if(set_socket_options(c->local_rfd.fd, 1)<0)
            longjmp(c->err, 1);
        auth_libwrap(c);
        auth_user(c);
        s_log(LOG_NOTICE, "%s connected from %s",
            c->opt->servname, c->accepting_address);
    }
}

static void init_remote(CLI *c) {
    /* create connection to host/service */
    if(c->opt->source_addr.num)
        memcpy(&c->bind_addr, &c->opt->source_addr, sizeof(SOCKADDR_LIST));
#ifndef USE_WIN32
    else if(c->opt->option.transparent)
        memcpy(&c->bind_addr, &c->peer_addr, sizeof(SOCKADDR_LIST));
#endif
    else {
        c->bind_addr.num=0; /* don't bind connecting socket */
    }

    /* setup c->remote_fd, now */
    if(c->opt->option.remote) {
        c->remote_fd.fd=connect_remote(c);
    } else /* NOT in remote mode */
        c->remote_fd.fd=connect_local(c);
    c->remote_fd.is_socket=1; /* Always! */
#ifndef USE_WIN32
    if(c->remote_fd.fd>=max_fds) {
        s_log(LOG_ERR, "Remote file descriptor out of range (%d>=%d)",
            c->remote_fd.fd, max_fds);
        longjmp(c->err, 1);
    }
#endif
    s_log(LOG_DEBUG, "Remote FD=%d initialized", c->remote_fd.fd);
    if(set_socket_options(c->remote_fd.fd, 2)<0)
        longjmp(c->err, 1);
}

static void init_ssl(CLI *c) {
#ifdef SSPI
    if (c->opt->option.client){
        handshake(c);
        c->cbwIoBuffer = 0;//set to zero lenght of write buffer after handshake
	}
    else{
        accept_handshake(c);
        c->cbwIoBuffer =0;//set to zero lenght of write buffer after handshake
	}
#else
    int i, err;
    SSL_SESSION *old_session;

    if(!(c->ssl=SSL_new(c->opt->ctx))) {
        sslerror("SSL_new");
        longjmp(c->err, 1);
    }
    SSL_set_ex_data(c->ssl, cli_index, c); /* for verify callback */
#if SSLEAY_VERSION_NUMBER >= 0x0922
    SSL_set_session_id_context(c->ssl, (unsigned char *)sid_ctx,
        strlen(sid_ctx));
#endif
    if(c->opt->option.client) {
        if(c->opt->session) {
            enter_critical_section(CRIT_SESSION);
            SSL_set_session(c->ssl, c->opt->session);
            leave_critical_section(CRIT_SESSION);
        }
        SSL_set_fd(c->ssl, c->remote_fd.fd);
        SSL_set_connect_state(c->ssl);
    } else {
        if(c->local_rfd.fd==c->local_wfd.fd)
            SSL_set_fd(c->ssl, c->local_rfd.fd);
        else {
           /* Does it make sence to have SSL on STDIN/STDOUT? */
            SSL_set_rfd(c->ssl, c->local_rfd.fd);
            SSL_set_wfd(c->ssl, c->local_wfd.fd);
        }
        SSL_set_accept_state(c->ssl);
    }
#endif // SSPI

    /* Setup some values for transfer() function */
    if(c->opt->option.client) {
        c->sock_rfd=&(c->local_rfd);
        c->sock_wfd=&(c->local_wfd);
        c->ssl_rfd=c->ssl_wfd=&(c->remote_fd);
    } else {
        c->sock_rfd=c->sock_wfd=&(c->remote_fd);
        c->ssl_rfd=&(c->local_rfd);
        c->ssl_wfd=&(c->local_wfd);
    }

#ifndef SSPI
    while(1) {
        if(c->opt->option.client)
            i=SSL_connect(c->ssl);
        else
            i=SSL_accept(c->ssl);
        err=SSL_get_error(c->ssl, i);
        if(err==SSL_ERROR_NONE)
            break; /* ok -> done */
        if(err==SSL_ERROR_WANT_READ || err==SSL_ERROR_WANT_WRITE) {
            s_poll_zero(&c->fds);
            s_poll_add(&c->fds, c->ssl_rfd->fd,
                err==SSL_ERROR_WANT_READ,
                err==SSL_ERROR_WANT_WRITE);
            switch(s_poll_wait(&c->fds, c->opt->timeout_busy, 0)) {
            case -1:
                sockerror("init_ssl: s_poll_wait");
                longjmp(c->err, 1);
            case 0:
                s_log(LOG_INFO, "init_ssl: s_poll_wait timeout");
                longjmp(c->err, 1);
            case 1:
                break; /* OK */
            default:
                s_log(LOG_ERR, "init_ssl: s_poll_wait unknown result");
                longjmp(c->err, 1);
            }
            continue; /* ok -> retry */
        }
        if(err==SSL_ERROR_SYSCALL) {
            switch(get_last_socket_error()) {
            case EINTR:
            case EAGAIN:
                continue;
            }
        }
        if(c->opt->option.client)
            sslerror("SSL_connect");
        else
            sslerror("SSL_accept");
        longjmp(c->err, 1);
    }
    if(SSL_session_reused(c->ssl)) {
        s_log(LOG_INFO, "SSL %s: previous session reused",
            c->opt->option.client ? "connected" : "accepted");
    } else { /* a new session was negotiated */
        if(c->opt->option.client) {
            s_log(LOG_INFO, "SSL connected: new session negotiated");
            enter_critical_section(CRIT_SESSION);
            old_session=c->opt->session;
            c->opt->session=SSL_get1_session(c->ssl); /* store it */
            if(old_session)
                SSL_SESSION_free(old_session); /* release the old one */
            leave_critical_section(CRIT_SESSION);
        } else
            s_log(LOG_INFO, "SSL accepted: new session negotiated");
        print_cipher(c);
    }
#endif //SSPI
}

/****************************** some defines for transfer() */
/* is socket/SSL open for read/write? */
#define sock_rd (c->sock_rfd->rd)
#define sock_wr (c->sock_wfd->wr)
#define ssl_rd  (c->ssl_rfd->rd)
#define ssl_wr  (c->ssl_wfd->wr)

/* is socket/SSL ready for read/write? */
#define sock_can_rd (s_poll_canread(&c->fds, c->sock_rfd->fd))
#define sock_can_wr (s_poll_canwrite(&c->fds, c->sock_wfd->fd))
#define ssl_can_rd  (s_poll_canread(&c->fds, c->ssl_rfd->fd))
#define ssl_can_wr  (s_poll_canwrite(&c->fds, c->ssl_wfd->fd))
/* NOTE: above defines are related to the logical data stream,
 * no longer to the underlying file descriptors */

/* does the underlaying file descriptor want to read/write? */
#ifndef SSPI
#define want_rd     (SSL_want_read(c->ssl))
#define want_wr     (SSL_want_write(c->ssl))
#else
#define SSL_shutdown(a)    SSPI_shutdown(a)
#define want_wr     c->sspi_want_wr
#define want_rd     c->sspi_want_rd 
#endif //SSPI
/****************************** transfer data */
static void transfer(CLI *c) {
    int num, err;
    int check_SSL_pending;
    int header_read = 1;
    enum {CL_OPEN, CL_INIT, CL_RETRY, CL_CLOSED} ssl_closing=CL_OPEN;
    int watchdog=0; /* a counter to detect an infinite loop */

    c->sock_ptr=c->ssl_ptr=0;
    sock_rd=sock_wr=ssl_rd=ssl_wr=1;
#ifdef SSPI    
    if (c->opt->tunnel_url)
    {
	header_read = 0;
	sprintf(c->sock_buff,"CONNECT %s HTTP/1.1\r\nHost: %s",
	    c->opt->tunnel_url,
	    c->opt->tunnel_url);
	if (strchr(c->opt->tunnel_url,':'))
	    *strrchr(c->sock_buff,':') = 0;
	strcat(c->sock_buff,"\r\n\r\n");    
	printf("String:%s\n",c->sock_buff);
	c->sock_ptr = strlen(c->sock_buff);
    }
#endif

    do { /* main loop */
        /* set flag to try and read any buffered SSL data
         * if we made room in the buffer by writing to the socket */
        check_SSL_pending=0;

        /****************************** setup c->fds structure */
        s_poll_zero(&c->fds); /* Initialize the structure */
        if(sock_rd && c->sock_ptr<BUFFSIZE) /* socket input buffer not full*/
            s_poll_add(&c->fds, c->sock_rfd->fd, 1, 0);
        if((ssl_rd && c->ssl_ptr<BUFFSIZE) || /* SSL input buffer not full */
                ((c->sock_ptr || ssl_closing==CL_RETRY) && want_rd)){
                /* want to SSL_write or SSL_shutdown but read from the
                 * underlying socket needed for the SSL protocol */
            s_log(LOG_DEBUG,"add ssl read socket to pool");
            s_log(LOG_DEBUG,"ssl_rd = %d, c->ssl_ptr = %d,c->sock_ptr=%d,want_rd = %d",ssl_rd,c->ssl_ptr,c->sock_ptr,want_rd);
            s_poll_add(&c->fds, c->ssl_rfd->fd, 1, 0);
        }
        if(c->ssl_ptr){ /* SSL input buffer not empty */
            s_log(LOG_DEBUG,"add write socket to poll");
            s_poll_add(&c->fds, c->sock_wfd->fd, 0, 1);
        }
        if(c->sock_ptr || /* socket input buffer not empty */
                ssl_closing==CL_INIT /* need to send close_notify */ ||
                ((c->ssl_ptr<BUFFSIZE || ssl_closing==CL_RETRY) && want_wr))
                /* want to SSL_read or SSL_shutdown but write to the
                 * underlying socket needed for the SSL protocol */
            s_poll_add(&c->fds, c->ssl_wfd->fd, 0, 1);

        /****************************** wait for an event */
#ifdef SSPI // do not need pool if there was extra data from previous call or from hanshake
if((!c->ExData && !c->sspi_buff_st)||c->ssl_ptr){
#endif 
        s_log(LOG_DEBUG,"Enter pool section on transfer");
        err=s_poll_wait(&c->fds, (sock_rd && ssl_rd) /* both peers open */ ||
            c->ssl_ptr /* data buffered to write to socket */ ||
            c->sock_ptr /* data buffered to write to SSL */ ?
            c->opt->timeout_idle : c->opt->timeout_close, 0);
        switch(err) {
        case -1:
            sockerror("transfer: s_poll_wait");
            longjmp(c->err, 1);
        case 0: /* timeout */
            if((sock_rd && ssl_rd) || c->ssl_ptr || c->sock_ptr) {
                s_log(LOG_INFO, "s_poll_wait timeout: connection reset");
                longjmp(c->err, 1);
            } else { /* already closing connection */
                s_log(LOG_INFO, "s_poll_wait timeout: connection close");
                return; /* OK */
            }
        }
        if(!(sock_can_rd || sock_can_wr || ssl_can_rd || ssl_can_wr)) {
            s_log(LOG_ERR, "INTERNAL ERROR: "
                "s_poll_wait returned %d, but no descriptor is ready", err);
            longjmp(c->err, 1);
        }

#ifdef SSPI
}
#endif
        /****************************** send SSL close_notify message */
        if(ssl_closing==CL_INIT || (ssl_closing==CL_RETRY &&
                ((want_rd && ssl_can_rd) || (want_wr && ssl_can_wr)))) {
            #ifndef SSPI
            switch(SSL_shutdown(c->ssl)) { /* Send close_notify */
            #endif //SSPI
            #ifdef SSPI
            switch(SSL_shutdown(c)) { /* Send close_notify */
            #endif //SSPI
            case 1: /* the shutdown was successfully completed */
                s_log(LOG_INFO, "SSL_shutdown successfully sent close_notify");
                ssl_rd = 0;
                ssl_wr = 0;
                ssl_closing=CL_CLOSED; /* done! */
                break;
            case 0: /* the shutdown is not yet finished */
                s_log(LOG_DEBUG, "SSL_shutdown retrying");
                ssl_closing=CL_RETRY; /* retry next time */
                break;
            case -1: /* a fatal error occurred */
#ifndef SSPI
                sslerror("SSL_shutdown");
#else
                s_log(LOG_ERR,"SSPI_shutdown error");
#endif //SSPI
                longjmp(c->err, 1);
            }
        }

        /****************************** write to socket */
        if(header_read && sock_wr && sock_can_wr) {
            num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
            switch(num) {
            case -1: /* error */
                parse_socket_error(c, "writesocket");
                break;
            case 0:
                s_log(LOG_DEBUG, "No data written to the socket: retrying");
                break;
            default:
                s_log(LOG_DEBUG, "data send to socket = %d",num);
                memmove(c->ssl_buff, c->ssl_buff+num, c->ssl_ptr-num);
                if(c->ssl_ptr>=BUFFSIZE) /* buffer was previously full */
                    check_SSL_pending=1; /* check for data buffered by SSL */
                c->ssl_ptr-=num;
                c->sock_bytes+=num;
		ios_run_watch_callback(0,num,0);
                watchdog=0; /* reset watchdog */
            }
        }

        /****************************** write to SSL */
        if((want_wr && ssl_can_wr) ||(ssl_wr && c->sock_ptr && ( /* output buffer not empty */
                ssl_can_wr || (want_rd && ssl_can_rd)
                /* SSL_write wants to read from the underlying descriptor */
                ))) {
#ifndef SSPI 
num=SSL_write(c->ssl, c->sock_buff, c->sock_ptr);
#else
num=SSPI_write(c);
#endif //SSPI
#ifndef SSPI
            switch(err=SSL_get_error(c->ssl, num)) {
            case SSL_ERROR_NONE:
                memmove(c->sock_buff, c->sock_buff+num, c->sock_ptr-num);
                c->sock_ptr-=num;
                c->ssl_bytes+=num;
		ios_run_watch_callback(0,0,num);
                watchdog=0; /* reset watchdog */
                break;
            case SSL_ERROR_WANT_WRITE:
                s_log(LOG_DEBUG, "SSL_write returned WANT_WRITE: retrying");
                break;
            case SSL_ERROR_WANT_READ:
                s_log(LOG_DEBUG, "SSL_write returned WANT_READ: retrying");
                break;
            case SSL_ERROR_WANT_X509_LOOKUP:
                s_log(LOG_DEBUG,
                    "SSL_write returned WANT_X509_LOOKUP: retrying");
                break;
            case SSL_ERROR_SYSCALL: /* really an error */
                if(num)
                    parse_socket_error(c, "SSL_write");
                break;
            case SSL_ERROR_ZERO_RETURN: /* close_notify received */
                s_log(LOG_DEBUG, "SSL closed on SSL_write");
                ssl_rd=0;
                break;
            case SSL_ERROR_SSL:
                sslerror("SSL_write");
                longjmp(c->err, 1);
            default:
                s_log(LOG_ERR, "SSL_write/SSL_get_error returned %d", err);
                longjmp(c->err, 1);
            }
#else //SSPI
        switch (num){
        case -1:
            s_log(LOG_ERR,"error on SSPI_write");
            longjmp(c->err,1);
            //break;
        case 0:
            s_log(LOG_NOTICE, "Zerro bytes send");
            break;
        default:
             s_log(LOG_DEBUG,"data send to ssl_socket =%d",num);
             memmove(c->sock_buff, c->sock_buff+num, c->sock_ptr-num);
             c->sock_ptr-=num;
             c->ssl_bytes+=num;
             watchdog=0; /* reset watchdog */
             break;
        }
#endif //SSPI
        }

        /****************************** read from socket */
        if(header_read && sock_rd && sock_can_rd) {
            num=readsocket(c->sock_rfd->fd,
                c->sock_buff+c->sock_ptr, BUFFSIZE-c->sock_ptr);
            switch(num) {
            case -1:
                parse_socket_error(c, "readsocket");
                break;
            case 0: /* close */
                s_log(LOG_DEBUG, "Socket closed on read");
                sock_rd=0;
                break;
            default:
                s_log(LOG_DEBUG,"data reciev from socket = %d",num);
                c->sock_ptr+=num;
                watchdog=0; /* reset watchdog */
            }
        }

        /****************************** read from SSL */
        if((ssl_rd && c->ssl_ptr<BUFFSIZE  && ssl_closing != CL_CLOSED &&( /* input buffer not full */
                ssl_can_rd || (want_wr && ssl_can_wr) ||
                /* SSL_read wants to write to the underlying descriptor */
#ifndef SSPI
(check_SSL_pending && SSL_pending(c->ssl))))) // it is ugly but in SSPI we need this 
#else
(check_SSL_pending && ssl_can_rd)))|| ((want_rd) && (ssl_can_rd || c->sspi_buff_st))||c->ExData) 
                /* write made space from full buffer */
#endif //SSPI 
            {
            #ifndef SSPI
            num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, BUFFSIZE-c->ssl_ptr);
            #else
            num=SSPI_read(c);
            #endif //SSPI
#ifndef SSPI
            switch(err=SSL_get_error(c->ssl, num)) {
            case SSL_ERROR_NONE:
                c->ssl_ptr+=num;
                watchdog=0; /* reset watchdog */
                break;
            case SSL_ERROR_WANT_WRITE:
                s_log(LOG_DEBUG, "SSL_read returned WANT_WRITE: retrying");
                break;
            case SSL_ERROR_WANT_READ:
                s_log(LOG_DEBUG, "SSL_read returned WANT_READ: retrying");
                break;
            case SSL_ERROR_WANT_X509_LOOKUP:
                s_log(LOG_DEBUG,
                    "SSL_read returned WANT_X509_LOOKUP: retrying");
                break;
            case SSL_ERROR_SYSCALL:
                if(!num) { /* EOF */
                    if(c->sock_ptr) {
                        s_log(LOG_ERR,
                            "SSL socket closed with %d byte(s) in buffer",
                            c->sock_ptr);
                        longjmp(c->err, 1); /* reset the socket */
                    }
                    s_log(LOG_DEBUG, "SSL socket closed on SSL_read");
                    ssl_rd=ssl_wr=0; /* buggy or SSLv2 peer: no close_notify */
                    ssl_closing=CL_CLOSED; /* don't try to send it back */
                } else
                    parse_socket_error(c, "SSL_read");
                break;
            case SSL_ERROR_ZERO_RETURN: /* close_notify received */
                s_log(LOG_DEBUG, "SSL closed on SSL_read");
                ssl_rd=0;
                break;
            case SSL_ERROR_SSL:
                sslerror("SSL_read");
                longjmp(c->err, 1);
            default:
                s_log(LOG_ERR, "SSL_read/SSL_get_error returned %d", err);
                longjmp(c->err, 1);
            }
#else //SSPI
        switch (num){
        case -3:
            s_log(LOG_NOTICE, "SSPI_read: read socket closed");
            ssl_rd=0;
            break;
        case -2:
            s_log(LOG_NOTICE, "SSPI_read return SEC_I_CONTEXT_EXPIRED");
            ssl_closing=CL_INIT; /* return SEC_I_CONTEXT_EXPIRED */
            ssl_rd=0; 
            break;
        case -1:
            s_log(LOG_ERR,"error on SSPI_read");
            longjmp(c->err,1);
            break;
        case 0:
            s_log(LOG_DEBUG, "Zerro bytes read");
            break;
        default:
             c->ssl_ptr+=num;
	    if (!header_read)
	    {
		char * ptr;
		c->ssl_buff[c->ssl_ptr]=0;
		ptr = strstr(c->ssl_buff,"\r\n\r\n");
		if (ptr)
		{
		    char * ptr2 = strstr(c->ssl_buff,"HTTP/1.1 200");
		    if (ptr2 == NULL || ptr2>ptr)
		    {
			s_log(LOG_ERR,"error on SSPI_read");
			longjmp(c->err,1);
		    }
		    c->ssl_ptr-=ptr-c->ssl_buff+4;
		    memmove(c->ssl_buff,ptr+4,c->ssl_ptr);
		    header_read=1;
		}
	    }

             s_log(LOG_DEBUG,"data read from ssl_sock =%d",num);
             watchdog=0; /* reset watchdog */
             break;
        }
#endif //SSPI

        }

        /****************************** check write shutdown conditions */
        if(sock_wr && !ssl_rd && !c->ssl_ptr) {
            s_log(LOG_DEBUG, "Socket write shutdown");
            s_log(LOG_DEBUG, "c->ssl_ptr = %d",c->ssl_ptr);
            sock_wr=0; /* no further write allowed */
            shutdown(c->sock_wfd->fd, SHUT_WR); /* send TCP FIN */
        }
#ifndef SSPI
	if(ssl_wr && (!sock_rd || SSL_get_shutdown(c->ssl)) && !c->sock_ptr) {
#else 
	if(ssl_wr && !sock_rd  && !c->sock_ptr && !want_wr) {
#endif //SSPI
            s_log(LOG_DEBUG, "SSL write shutdown");
            ssl_wr=0; /* no further write allowed */
#ifndef SSPI
	    if(strcmp(SSL_get_version(c->ssl), "SSLv2")) { /* SSLv3, TLSv1 */
#endif
                ssl_closing=CL_INIT; /* initiate close_notify */
#ifndef SSPI
            } else { /* no alerts in SSLv2 including close_notify alert */
                shutdown(c->sock_rfd->fd, SHUT_RD); /* notify the kernel */
                shutdown(c->sock_wfd->fd, SHUT_WR); /* send TCP FIN */
                SSL_set_shutdown(c->ssl, /* notify the OpenSSL library */
                    SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                ssl_rd=0; /* no further read allowed */
                ssl_closing=CL_CLOSED; /* closed */
            }
#endif //SSPI
        }
        if(ssl_closing==CL_RETRY) { /* SSL shutdown */
            if(!want_rd && !want_wr) { /* close_notify alert was received */
                s_log(LOG_DEBUG, "SSL doesn't need to read or write");
                ssl_closing=CL_CLOSED;
            }
            if(watchdog>5) {
                s_log(LOG_NOTICE, "Too many retries on SSL shutdown");
                ssl_closing=CL_CLOSED;
            }
        }

        /****************************** check watchdog */
        if(++watchdog>100) { /* loop executes without transferring any data */
            s_log(LOG_ERR,
                "transfer() loop executes not transferring any data");
            s_log(LOG_ERR,
                "please report the problem to Michal.Trojnara@mirt.net");
            s_log(LOG_ERR, "socket open: rd=%s wr=%s, ssl open: rd=%s wr=%s",
                sock_rd ? "yes" : "no", sock_wr ? "yes" : "no",
                ssl_rd ? "yes" : "no", ssl_wr ? "yes" : "no");
            s_log(LOG_ERR, "socket ready: rd=%s wr=%s, ssl ready: rd=%s wr=%s",
                sock_can_rd ? "yes" : "no", sock_can_wr ? "yes" : "no",
                ssl_can_rd ? "yes" : "no", ssl_can_wr ? "yes" : "no");
            s_log(LOG_ERR, "ssl want: rd=%s wr=%s",
                want_rd ? "yes" : "no", want_wr ? "yes" : "no");
            s_log(LOG_ERR, "socket input buffer: %d byte(s), "
                "ssl input buffer: %d byte(s)", c->sock_ptr, c->ssl_ptr);
            s_log(LOG_ERR, "check_SSL_pending=%d, ssl_closing=%d",
                check_SSL_pending, ssl_closing);
            longjmp(c->err, 1);
        }

    } while(sock_wr || ssl_closing!=CL_CLOSED);
}

static void parse_socket_error(CLI *c, const char *text) {
    switch(get_last_socket_error()) {
    case EINTR:
        s_log(LOG_DEBUG, "%s interrupted by a signal: retrying", text);
        return;
    case EWOULDBLOCK:
        s_log(LOG_NOTICE, "%s would block: retrying", text);
        sleep(1); /* Microsoft bug KB177346 */
        return;
#if EAGAIN!=EWOULDBLOCK
    case EAGAIN:
        s_log(LOG_DEBUG, "%s temporary lack of resources: retrying", text);
        return;
#endif
    default:
        sockerror(text);
        longjmp(c->err, 1);
    }
}

#ifndef SSPI
static void print_cipher(CLI *c) { /* print negotiated cipher */
#if SSLEAY_VERSION_NUMBER <= 0x0800
    s_log(LOG_INFO, "%s opened with SSLv%d, cipher %s",
        c->opt->servname, ssl->session->ssl_version, SSL_get_cipher(c->ssl));
#else
    SSL_CIPHER *cipher;
    char buf[STRLEN];
    int len;

    cipher=SSL_get_current_cipher(c->ssl);
    SSL_CIPHER_description(cipher, buf, STRLEN);
    len=strlen(buf);
    if(len>0)
        buf[len-1]='\0';
    s_log(LOG_INFO, "Negotiated ciphers: %s", buf);
#endif
}
#endif //SSPI

static void auth_libwrap(CLI *c) {
#ifdef USE_LIBWRAP
    struct request_info request;
    int fd[2];
    int result=0; /* deny by default */

    if(pipe(fd)<0) {
        ioerror("pipe");
        longjmp(c->err, 1);
    }
    if(alloc_fd(fd[0]) || alloc_fd(fd[1]))
        longjmp(c->err, 1);
    switch(fork()) {
    case -1:    /* error */
        close(fd[0]);
        close(fd[1]);
        ioerror("fork");
        longjmp(c->err, 1);
    case  0:    /* child */
        close(fd[0]); /* read side */
        request_init(&request,
            RQ_DAEMON, c->opt->servname, RQ_FILE, c->local_rfd.fd, 0);
        fromhost(&request);
        result=hosts_access(&request);
        write_blocking(c, fd[1], (u8 *)&result, sizeof(result));
            /* ignore the returned error */
        close(fd[1]); /* write side */
        _exit(0);
    default:    /* parent */
        close(fd[1]); /* write side */
        read_blocking(c, fd[0], (u8 *)&result, sizeof(result));
            /* ignore the returned error */
        close(fd[0]); /* read side */
        /* no need to wait() for zombies here:
         *  - in UCONTEXT/PTHREAD mode they're removed using the signal pipe
         *  - in FORK mode they're removed with the client process */
    }

    if(!result) {
        s_log(LOG_WARNING, "Connection from %s REFUSED by libwrap",
            c->accepting_address);
        s_log(LOG_DEBUG, "See hosts_access(5) manual for details");
        longjmp(c->err, 1);
    }
    s_log(LOG_DEBUG, "Connection from %s permitted by libwrap",
        c->accepting_address);
#endif
}

static void auth_user(CLI *c) {
#ifndef HSM_SERVER // no use of it currently
#ifndef _WIN32_WCE
    struct servent *s_ent;    /* structure for getservbyname */
#endif
    SOCKADDR_UNION ident;     /* IDENT socket name */
    char name[STRLEN];
    int error;

    if(!c->opt->username)
        return; /* -u option not specified */
    if((c->fd=
            socket(c->peer_addr.addr[0].sa.sa_family, SOCK_STREAM, 0))<0) {
        sockerror("socket (auth_user)");
        longjmp(c->err, 1);
    }
    if(alloc_fd(c->fd))
        longjmp(c->err, 1);
    memcpy(&ident, &c->peer_addr.addr[0], sizeof(SOCKADDR_UNION));
#ifndef _WIN32_WCE
    s_ent=getservbyname("auth", "tcp");
    if(s_ent) {
        ident.in.sin_port=s_ent->s_port;
    } else
#endif
    {
        s_log(LOG_WARNING, "Unknown service 'auth': using default 113");
        ident.in.sin_port=htons(113);
    }
    if(st_connect(c->fd, &ident.sa, addr_len(ident))) {
        error=get_last_socket_error();
        if(error!=EINPROGRESS && error!=EWOULDBLOCK) {
            sockerror("ident connect (auth_user)");
            longjmp(c->err, 1);
        }
        if(connect_wait(c))
            longjmp(c->err, 1);
    }
    s_log(LOG_DEBUG, "IDENT server connected");
    fdprintf(c, c->fd, "%u , %u",
        ntohs(c->peer_addr.addr[0].in.sin_port),
        ntohs(c->opt->local_addr.addr[0].in.sin_port));
    if(fdscanf(c, c->fd, "%*[^:]: USERID :%*[^:]:%s", name)!=1) {
        s_log(LOG_ERR, "Incorrect data from IDENT server");
        longjmp(c->err, 1);
    }
    closesocket(c->fd);
    c->fd=-1; /* avoid double close on cleanup */
    if(strcmp(name, c->opt->username)) {
        safestring(name);
        s_log(LOG_WARNING, "Connection from %s REFUSED by IDENT (user %s)",
            c->accepting_address, name);
        longjmp(c->err, 1);
    }
    s_log(LOG_INFO, "IDENT authentication passed");
#endif //HSM_SERVER    
}

static int connect_local(CLI *c) { /* spawn local process */
#if defined (USE_WIN32) || defined (__vms)
    s_log(LOG_ERR, "LOCAL MODE NOT SUPPORTED ON WIN32 and OpenVMS PLATFORM");
    longjmp(c->err, 1);
    return -1; /* some C compilers require a return value */
#else /* USE_WIN32, __vms */
    char env[3][STRLEN],  *portname;
    int fd[2], pid;
#ifndef SSPI
    char name[STRLEN];
    X509 *peer;
#endif //SSPI
#ifdef HAVE_PTHREAD_SIGMASK
    sigset_t newmask;
#endif

    if (c->opt->option.pty) {
        char tty[STRLEN];

        if(pty_allocate(fd, fd+1, tty, STRLEN))
            longjmp(c->err, 1);
        s_log(LOG_DEBUG, "%s allocated", tty);
    } else
        make_sockets(c, fd);
    pid=fork();
    c->pid=(unsigned long)pid;
    switch(pid) {
    case -1:    /* error */
        closesocket(fd[0]);
        closesocket(fd[1]);
        ioerror("fork");
        longjmp(c->err, 1);
    case  0:    /* child */
        closesocket(fd[0]);
        dup2(fd[1], 0);
        dup2(fd[1], 1);
        if(!options.option.foreground)
            dup2(fd[1], 2);
        closesocket(fd[1]);
        safecopy(env[0], "REMOTE_HOST=");
        safeconcat(env[0], c->accepting_address);
        portname=strrchr(env[0], ':');
        if(portname) /* strip the port name */
            *portname='\0';
        putenv(env[0]);
        if(c->opt->option.transparent) {
            putenv("LD_PRELOAD=" LIBDIR "/libstunnel.so");
            /* For Tru64 _RLD_LIST is used instead */
            putenv("_RLD_LIST=" LIBDIR "/libstunnel.so:DEFAULT");
        }
#ifndef SSPI
        if(c->ssl) {
            peer=SSL_get_peer_certificate(c->ssl);
            if(peer) {
                safecopy(env[1], "SSL_CLIENT_DN=");
                X509_NAME_oneline(X509_get_subject_name(peer), name, STRLEN);
                safestring(name);
                safeconcat(env[1], name);
                putenv(env[1]);
                safecopy(env[2], "SSL_CLIENT_I_DN=");
                X509_NAME_oneline(X509_get_issuer_name(peer), name, STRLEN);
                safestring(name);
                safeconcat(env[2], name);
                putenv(env[2]);
                X509_free(peer);
            }
        }
#endif //SSPI
#ifdef HAVE_PTHREAD_SIGMASK
        sigemptyset(&newmask);
        sigprocmask(SIG_SETMASK, &newmask, NULL);
#endif
        execvp(c->opt->execname, c->opt->execargs);
        ioerror(c->opt->execname); /* execv failed */
        _exit(1);
    default:
        break;
    }
    /* parent */
    s_log(LOG_INFO, "Local mode child started (PID=%lu)", c->pid);
    closesocket(fd[1]);
#ifdef FD_CLOEXEC
    fcntl(fd[0], F_SETFD, FD_CLOEXEC);
#endif
    return fd[0];
#endif /* USE_WIN32,__vms */
}

#ifndef USE_WIN32

static void make_sockets(CLI *c, int fd[2]) { /* make a pair of connected sockets */
#ifdef INET_SOCKET_PAIR
    SOCKADDR_UNION addr;
    socklen_t addrlen;
    int s; /* temporary socket awaiting for connection */

    if((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
        sockerror("socket#1");
        longjmp(c->err, 1);
    }
    if((fd[1]=socket(AF_INET, SOCK_STREAM, 0))<0) {
        sockerror("socket#2");
        longjmp(c->err, 1);
    }
    addrlen=sizeof(SOCKADDR_UNION);
    memset(&addr, 0, addrlen);
    addr.in.sin_family=AF_INET;
    addr.in.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
    addr.in.sin_port=0; /* dynamic port allocation */
    if(bind(s, &addr.sa, addrlen))
        log_error(LOG_DEBUG, get_last_socket_error(), "bind#1");
    if(bind(fd[1], &addr.sa, addrlen))
        log_error(LOG_DEBUG, get_last_socket_error(), "bind#2");
    if(st_listen(s, 5)) {
        sockerror("listen");
        longjmp(c->err, 1);
    }
    if(getsockname(s, &addr.sa, &addrlen)) {
        sockerror("getsockname");
        longjmp(c->err, 1);
    }
    if(st_connect(fd[1], &addr.sa, addrlen)) {
        sockerror("connect");
        longjmp(c->err, 1);
    }
    if((fd[0]=st_accept(s, &addr.sa, &addrlen))<0) {
        sockerror("accept");
        longjmp(c->err, 1);
    }
    closesocket(s); /* don't care about the result */
#else
    if(socketpair(AF_UNIX, SOCK_STREAM, 0, fd)) {
        sockerror("socketpair");
        longjmp(c->err, 1);
    }
#endif
}
#endif

static int connect_remote(CLI *c) { /* connect to remote host */
    SOCKADDR_UNION bind_addr, addr;
    SOCKADDR_LIST resolved_list, *address_list;
    int error, fd;
    u16 i;
#ifndef _WIN32
    if(*c->opt->remote_address == '/') {
	resolved_list.num = 1;
	resolved_list.cur = 0;
	resolved_list.addr[0].un.sun_family = AF_UNIX;
	strncpy (resolved_list.addr[0].un.sun_path, 
	    c->opt->remote_address, 
	    sizeof (resolved_list.addr[0].un.sun_path)-1);
	resolved_list.addr[0].un.sun_path [ sizeof (resolved_list.addr[0].un.sun_path)-1] = 0;
        address_list=&resolved_list;
    } else
#endif //_WIN32
    /* setup address_list */
    if(c->opt->option.delayed_lookup) {
        resolved_list.num=0;
        if(!name2addrlist(&resolved_list,
                c->opt->remote_address, DEFAULT_LOOPBACK)){
            s_log(LOG_ERR, "No host resolved");
            longjmp(c->err, 1);
        }
        address_list=&resolved_list;
    } else /* use pre-resolved addresses */
        address_list=&c->opt->remote_addr;

    /* try to connect each host from the list */
    for(i=0; i<address_list->num; i++) {
        memcpy(&addr, address_list->addr + address_list->cur,
            sizeof(SOCKADDR_UNION));
        address_list->cur=(address_list->cur+1)%address_list->num;
        /* race condition is possible, but harmless in this case */

        if((c->fd=socket(addr.sa.sa_family, SOCK_STREAM, 0))<0) {
            sockerror("remote socket");
            longjmp(c->err, 1);
        }
        if(alloc_fd(c->fd))
            longjmp(c->err, 1);
#ifndef USE_WIN32
	if(addr.sa.sa_family == AF_UNIX)
	    strncpy(c->connecting_address, addr.un.sun_path, sizeof (c->connecting_address));
	else
#endif //USE_WIN32
        if(c->bind_addr.num) { /* explicit local bind or transparent proxy */
            memcpy(&bind_addr, &c->bind_addr.addr[0], sizeof(SOCKADDR_UNION));
            if(bind(c->fd, &bind_addr.sa, addr_len(bind_addr))<0) {
                sockerror("bind transparent");
                longjmp(c->err, 1);
            }
	    s_ntop(c->connecting_address, &addr);
        }

        /* try to connect for the 1st time */
        s_log(LOG_DEBUG, "%s connecting %s",
            c->opt->servname, c->connecting_address);
        if(!st_connect(c->fd, &addr.sa, addr_len(addr))) {
#ifndef USE_WIN32
	    if(options.option.for_hsm && addr.sa.sa_family == AF_UNIX){
#if defined HAVE_PASSCRED && defined LINUX
	      int yesno = 1;
	      if (setsockopt (c->fd, SOL_SOCKET, SO_PASSCRED, (LPCSTR)&yesno, sizeof(yesno))) {
		s_log(LOG_ERR, " setsockopt(SO_PASSCRED) fail: 0x%lx", (unsigned long)errno);
	      }
#endif
	      if(use_passcred1(c))
                longjmp(c->err, 1);
        }
#endif
            fd=c->fd;
            c->fd=-1;
            return fd; /* no error -> success (should not be possible) */
        }
        error=get_last_socket_error();
        if(error!=EINPROGRESS && error!=EWOULDBLOCK) {
            s_log(LOG_ERR, "remote connect (%s): %s (%d)",
                c->connecting_address, my_strerror(error), error);
            closesocket(c->fd);
            c->fd=-1;
            continue; /* next IP */
        }
        if(connect_wait(c))
            longjmp(c->err, 1);
#ifndef USE_WIN32
	if(options.option.for_hsm && addr.sa.sa_family == AF_UNIX)
	  if(use_passcred1(c))
            longjmp(c->err, 1);
#endif //USE_WIN32
        fd=c->fd;
        c->fd=-1;
        return fd; /* success! */
    }
    longjmp(c->err, 1);
    return -1; /* some C compilers require a return value */
}

    /* wait for the result of a non-blocking connect */
    /* file descriptor : c->fd                       */
    /* timeout         : c->opt->timeout_connect     */
int connect_wait(CLI *c) {
    int error;
    socklen_t optlen;

    s_log(LOG_DEBUG, "connect_wait: waiting %d seconds",
        c->opt->timeout_connect);
    s_poll_zero(&c->fds);
    s_poll_add(&c->fds, c->fd, 1, 1);
    switch(s_poll_wait(&c->fds, c->opt->timeout_connect, 0)) {
    case -1:
        sockerror("connect_wait: s_poll_wait");
        return -1;
    case 0:
        s_log(LOG_INFO, "connect_wait: s_poll_wait timeout");
        return -1;
    default:
        if(s_poll_canread(&c->fds, c->fd)) {
            /* just connected socket should not be ready for read */
            /* get the resulting error code, now */
            optlen=sizeof(error);
            if(!getsockopt(c->fd, SOL_SOCKET, SO_ERROR, (void *)&error, &optlen))
                errno=error;
            if(errno) { /* really an error? */
                sockerror("connect_wait: getsockopt");
                return -1;
            }
        }
        if(s_poll_canwrite(&c->fds, c->fd)) {
            s_log(LOG_DEBUG, "connect_wait: connected");
            return 0; /* success */
        }
        s_log(LOG_ERR, "connect_wait: unexpected s_poll_wait result");
        return -1;
    }
#ifndef SOLARIS
    return -1; /* should not be possible */
#endif
}

static void reset(int fd, char *txt) {
    /* Set lingering on a socket if needed*/
    struct linger l;

    l.l_onoff=1;
    l.l_linger=0;
    if(setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *)&l, sizeof(l)))
        log_error(LOG_DEBUG, get_last_socket_error(), txt);
}
#ifdef SSPI

//-------------------------------------------------------------

SECURITY_STATUS
CreateClientCredentials(
    CLI* c)            // out
{
    TimeStamp       tsExpiry;
    SECURITY_STATUS Status;
    PCCERT_CONTEXT  pCert = NULL;
    SCHANNEL_CRED   SchannelCred;
    PCCERT_CONTEXT  pCertContext = NULL;

    if(c->opt->option.cert){
        s_log(LOG_NOTICE,"try to read the client certificate");
        get_cert_from_file(c->opt->cert, &pCert);
        if(!pCert){
            s_log(LOG_ERR,"CertCreateCertificateContext failed: %ud\n", GetLastError( ) );
            return SEC_E_NO_CREDENTIALS;
	}
	pCertContext = CertFindCertificateInStore(c->hMyCertStore, 
                                              X509_ASN_ENCODING|PKCS_7_ASN_ENCODING, 
                                              0,
                                              CERT_FIND_EXISTING,
                                              pCert,
                                              NULL);
	if(pCertContext == NULL){
            s_log(LOG_NOTICE,"CertFindCertificateInStore not find client certificate in store CURRENT_USER. Looking at LOCAL_MACHINE ");
            pCertContext = CertFindCertificateInStore(c->hSysCertStore, 
                                                  X509_ASN_ENCODING|PKCS_7_ASN_ENCODING, 
                                                  0,
                                                  CERT_FIND_EXISTING,
                                                  pCert,
                                                  NULL);

        }
	CertFreeCertificateContext(pCert);
        if(pCertContext == NULL){
            s_log(LOG_ERR,"Error 0x%x returned by CertFindCertificateInStore\n",
            GetLastError());
            return SEC_E_NO_CREDENTIALS;
        }
    }
    if(c->opt->pincode){
        DWORD keytype = 0;
        BOOL should_release_ctx = FALSE;
        s_log(LOG_NOTICE,"pincode option is present. Call CryptSetProvParam");
        Status = CryptAcquireCertificatePrivateKey(pCertContext,0,NULL,&c->opt->hProv,&keytype,&should_release_ctx);
        if(!Status){
            s_log(LOG_ERR,"CryptAcquireCertificatePrivateKey failed. Error = 0x%x", GetLastError());
            return SEC_E_NO_CREDENTIALS;
        }
        Status = CryptSetProvParam(c->opt->hProv,PP_KEYEXCHANGE_PIN,
	    		   (BYTE*)c->opt->pincode,0);
        if(!Status){
            BYTE buff[255];
            DWORD len = sizeof(buff);
            s_log(LOG_ERR,"CryptSetProvParam PP_KEYEXCHANGE_PIN failed. Error = 0x%x, pincode=%s", GetLastError(),c->opt->pincode);
            CryptGetProvParam(c->opt->hProv,PP_CONTAINER,buff,&len,0);
            s_log(LOG_ERR,"Container Name = %s",(char*)buff);
            return SEC_E_NO_CREDENTIALS;
        }
        s_log(LOG_NOTICE,"Call CryptSetProvParam(change ping) OK");
    }
    ZeroMemory(&SchannelCred, sizeof(SchannelCred));
    SchannelCred.dwVersion  = SCHANNEL_CRED_VERSION;
    if(pCertContext){
        SchannelCred.cCreds     = 1;
        SchannelCred.paCred     = &pCertContext;
    }
    SchannelCred.grbitEnabledProtocols = dwProtocol;
    SchannelCred.dwFlags |= SCH_CRED_NO_DEFAULT_CREDS;
    SchannelCred.dwFlags |= SCH_CRED_MANUAL_CRED_VALIDATION;

    Status = g_pSSPI->AcquireCredentialsHandle(
                        NULL,                   // Name of principal    
                        UNISP_NAME,           // Name of package
                        SECPKG_CRED_OUTBOUND,   // Flags indicating use
                        NULL,                   // Pointer to logon ID
                        &SchannelCred,          // Package specific data
                        NULL,                   // Pointer to GetKey() func
                        NULL,                   // Value to pass to GetKey()
			&c->hClientCreds,                // (out) Cred Handle
                        &tsExpiry);             // (out) Lifetime (optional)
    if(Status != SEC_E_OK)
    {
        s_log(LOG_ERR,"**** Error 0x%x returned by AcquireCredentialsHandle", Status);
        CertFreeCertificateContext(pCertContext);
    }
        s_log(LOG_ERR,"Credentials complete");
    CertFreeCertificateContext(pCertContext);
    return Status;
} //CreateCredentials


SECURITY_STATUS
PerformClientHandshake(
    SOCKET          Socket,         // in
    PCredHandle     phCreds,        // in
    LPSTR           pszServerName,  // in
    CtxtHandle *    phContext,      // out
    SecBuffer *     pExtraData,
    CLI*            c)     // out
{
    SecBufferDesc   OutBuffer;
    SecBuffer       OutBuffers[1];
    DWORD           dwSSPIFlags;
    unsigned long   dwSSPIOutFlags;
    TimeStamp       tsExpiry;
    SECURITY_STATUS scRet;
    int		    cbData;

    dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT   |
                  ISC_REQ_REPLAY_DETECT     |
                  ISC_REQ_CONFIDENTIALITY   |
                  ISC_RET_EXTENDED_ERROR    |
                  ISC_REQ_ALLOCATE_MEMORY   |
		  ISC_REQ_STREAM;


    OutBuffers[0].pvBuffer   = NULL;
    OutBuffers[0].BufferType = SECBUFFER_TOKEN;
    OutBuffers[0].cbBuffer   = 0;

    OutBuffer.cBuffers = 1;
    OutBuffer.pBuffers = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;
#ifdef UNICODE
    {
    WCHAR* pszwServerName = str2tstr(pszServerName);
#endif
    scRet = g_pSSPI->InitializeSecurityContext(
                    phCreds,
                    NULL,
#ifdef UNICODE
                    pszwServerName,
#else
		    pszServerName,
#endif
                    dwSSPIFlags,
                    0,
                    SECURITY_NATIVE_DREP,
                    NULL,
                    0,
                    phContext,
                    &OutBuffer,
                    &dwSSPIOutFlags,
                    &tsExpiry);
#ifdef UNICODE
    free(pszwServerName);
    }
#endif
    if(scRet != SEC_I_CONTINUE_NEEDED)
    {
        s_log(LOG_ERR,"**** Error %d returned by InitializeSecurityContext (1)", scRet);
        return scRet;
    }

    if(OutBuffers[0].cbBuffer != 0 && OutBuffers[0].pvBuffer != NULL)
    {
    cbData = send(Socket,
                  OutBuffers[0].pvBuffer,
                  OutBuffers[0].cbBuffer,
                  0);
    if(IS_SOCKET_ERROR(cbData) || cbData == 0)
    {
        s_log(LOG_ERR,"**** Error %d sending data to server (1)", WSAGetLastError());
        g_pSSPI->FreeContextBuffer(OutBuffers[0].pvBuffer);
        g_pSSPI->DeleteSecurityContext(phContext);
        return SEC_E_INTERNAL_ERROR;
    }

    s_log(LOG_DEBUG,"%d bytes of handshake data sent", cbData);

    
    g_pSSPI->FreeContextBuffer(OutBuffers[0].pvBuffer);
    OutBuffers[0].pvBuffer = NULL;
    }


    return ClientHandshakeLoop(Socket, phCreds, phContext, TRUE, pExtraData, c);
} // PerformClientHandshake

//-------------------------------------------------------------

SECURITY_STATUS
ClientHandshakeLoop(
    SOCKET          Socket,         // in
    PCredHandle     phCreds,        // in
    CtxtHandle *    phContext,      // in, out
    BOOL            fDoInitialRead, // in
    SecBuffer *     pExtraData,
    CLI*            c)
{
    SecBufferDesc   InBuffer;
    SecBuffer       InBuffers[2];
    SecBufferDesc   OutBuffer;
    SecBuffer       OutBuffers[1];
    DWORD           dwSSPIFlags;
    unsigned long   dwSSPIOutFlags;
    TimeStamp       tsExpiry;
    SECURITY_STATUS scRet;
    int		    cbData =0;
    int             wRes;
    s_poll_set fds; /* File descriptors */
    BOOL            fDoRead;


    dwSSPIFlags = ISC_REQ_SEQUENCE_DETECT   |
                  ISC_REQ_REPLAY_DETECT     |
                  ISC_REQ_CONFIDENTIALITY   |
                  ISC_RET_EXTENDED_ERROR    |
                  ISC_REQ_ALLOCATE_MEMORY   |
                  ISC_REQ_STREAM;


    c->cbwIoBuffer = 0;
    fDoRead = fDoInitialRead;



    scRet = SEC_I_CONTINUE_NEEDED;
    s_poll_zero(&fds);
    s_poll_add(&fds,Socket,1,0);//add for s_poll_wait
    while(scRet == SEC_I_CONTINUE_NEEDED        ||
          scRet == SEC_E_INCOMPLETE_MESSAGE     ||
          scRet == SEC_I_INCOMPLETE_CREDENTIALS) 
  { 
        if(0 == c->cbwIoBuffer || scRet == SEC_E_INCOMPLETE_MESSAGE)
        {
            if(fDoRead)
            
            {
                wRes = s_poll_wait(&fds,200,0);
                switch(wRes) {
                    case -1:
                        sockerror("hanshake_loop: s_poll_wait error");
                        return -1;
                    case 0: /* timeout */
                        s_log(LOG_INFO, "handshake loop s_poll_wait timeout: connection reset");
                        return -1;
                }
                cbData = recv(Socket, 
                     c->wIoBuffer + c->cbwIoBuffer, 
                     BUFFSIZE - c->cbwIoBuffer, 
                     0);
                if(IS_SOCKET_ERROR(cbData) && errno != EAGAIN)
                {
                    s_log(LOG_ERR,"**** Error %d reading data from server", WSAGetLastError());
                    scRet = SEC_E_INTERNAL_ERROR;
                    return scRet;
                }
                else if(cbData == 0 && errno != EAGAIN)
                {
                    s_log(LOG_ERR,"**** Server unexpectedly disconnected");
                    scRet = SEC_E_INTERNAL_ERROR;
                    return scRet;
                }
            {
                c->cbwIoBuffer += cbData;
                s_log(LOG_NOTICE,"%d bytes of handshake(in handshake loop) data received.", cbData);
            }
            }
            else
            {
                fDoRead = TRUE;
            }
        }

        InBuffers[0].pvBuffer   = c->wIoBuffer;
        InBuffers[0].cbBuffer   = c->cbwIoBuffer;
        InBuffers[0].BufferType = SECBUFFER_TOKEN;

        InBuffers[1].pvBuffer   = NULL;
        InBuffers[1].cbBuffer   = 0;
        InBuffers[1].BufferType = SECBUFFER_EMPTY;

        InBuffer.cBuffers       = 2;
        InBuffer.pBuffers       = InBuffers;
        InBuffer.ulVersion      = SECBUFFER_VERSION;


        OutBuffers[0].pvBuffer  = NULL;
        OutBuffers[0].BufferType= SECBUFFER_TOKEN;
        OutBuffers[0].cbBuffer  = 0;

        OutBuffer.cBuffers      = 1;
        OutBuffer.pBuffers      = OutBuffers;
        OutBuffer.ulVersion     = SECBUFFER_VERSION;


        scRet = g_pSSPI->InitializeSecurityContext(phCreds,
                                          phContext,
                                          NULL,
                                          dwSSPIFlags,
                                          0,
                                          SECURITY_NATIVE_DREP,
                                          &InBuffer,
                                          0,
                                          NULL,
                                          &OutBuffer,
                                          &dwSSPIOutFlags,
                                          &tsExpiry);

        if(scRet == SEC_E_OK                ||
           scRet == SEC_I_CONTINUE_NEEDED   ||(
           FAILED(scRet) && (dwSSPIOutFlags & ISC_RET_EXTENDED_ERROR)))
        {
            if(OutBuffers[0].cbBuffer != 0 && OutBuffers[0].pvBuffer != NULL)
            {
                cbData = send(Socket,
                              OutBuffers[0].pvBuffer,
                              OutBuffers[0].cbBuffer,
                              0);
                if(IS_SOCKET_ERROR(cbData) || cbData == 0)
                {
                    s_log(LOG_ERR,"**** Error %d sending data to server (2)", 
                        WSAGetLastError());
                    g_pSSPI->FreeContextBuffer(OutBuffers[0].pvBuffer);
                    g_pSSPI->DeleteSecurityContext(phContext);
                    return SEC_E_INTERNAL_ERROR;
                }

                s_log(LOG_NOTICE,"%d bytes of handshake data sent", cbData);

               
                g_pSSPI->FreeContextBuffer(OutBuffers[0].pvBuffer);
                OutBuffers[0].pvBuffer = NULL;
            }
        }

        if(scRet == SEC_E_INCOMPLETE_MESSAGE){
            continue;
        }
        if(scRet == SEC_E_OK)
        {
            s_log(LOG_NOTICE,"Handshake was successful");

            if(InBuffers[1].BufferType == SECBUFFER_EXTRA)
            {
                c->extra_data++;
                MoveMemory(c->ExDataBuffer,
                           c->wIoBuffer + (c->cbwIoBuffer - InBuffers[1].cbBuffer),
                           InBuffers[1].cbBuffer);
                c->ExLen  = InBuffers[1].cbBuffer;
                c->sspi_buff_st = 1;
                c->ExData = 1;
                s_log(LOG_NOTICE,"%d bytes of app data was bundled with handshake data",
                pExtraData->cbBuffer);
            }
            else
            {
                pExtraData->pvBuffer   = NULL;
                pExtraData->cbBuffer   = 0;
                pExtraData->BufferType = SECBUFFER_EMPTY;
            }
            break;
        }
        if(FAILED(scRet)){
            s_log(LOG_ERR,"**** Error 0x%x returned by InitializeSecurityContext (2)", scRet);
            break;
        }
        if(scRet == SEC_I_INCOMPLETE_CREDENTIALS){
            GetNewClientCredentials(c,phCreds, phContext);
            fDoRead = FALSE;
            scRet = SEC_I_CONTINUE_NEEDED;
	    //   Platform SDK!
	    // ,                    
            c->cbwIoBuffer = 0;
            continue;
        }
        if ( InBuffers[1].BufferType == SECBUFFER_EXTRA ){
            c->extra_data++;
            MoveMemory(c->wIoBuffer,
                       c->wIoBuffer + (c->cbwIoBuffer - InBuffers[1].cbBuffer),
                       InBuffers[1].cbBuffer);
            c->cbwIoBuffer = InBuffers[1].cbBuffer;
        }
        else
        {
            c->cbwIoBuffer = 0;
        }
    }

    if(FAILED(scRet)){
        g_pSSPI->DeleteSecurityContext(phContext);
    }


    return scRet;
} // ClientHandshakeLoop

 
void
DisplayCertChain(
    PCCERT_CONTEXT  pServerCert,
    BOOL            fLocal)
{
#if defined UNICODE
    WCHAR szName[1000];
#else
    CHAR szName[1000];
#endif
    PCCERT_CONTEXT pCurrentCert;
    PCCERT_CONTEXT pIssuerCert;
    DWORD dwVerificationFlags;


    if(!CertNameToStr(pServerCert->dwCertEncodingType,
                      &pServerCert->pCertInfo->Subject,
                      CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
                      szName, sizeof(szName)))
    {
        s_log(LOG_ERR,"**** Error 0x%x building subject name", GetLastError());
    }
    if(fLocal)
    {
        s_log(LOG_NOTICE,"Client subject: %s", szName);
    }
    else
    {
        s_log(LOG_NOTICE,"Server subject: %s", szName);
    }
    if(!CertNameToStr(pServerCert->dwCertEncodingType,
                      &pServerCert->pCertInfo->Issuer,
                      CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
                      szName, sizeof(szName)))
    {
        s_log(LOG_ERR,"**** Error 0x%x building issuer name", GetLastError());
    }
    if(fLocal)
    {
        s_log(LOG_NOTICE,"Client issuer: %s", szName);
    }
    else
    {
        s_log(LOG_NOTICE,"Server issuer: %s", szName);
    }


    pCurrentCert = pServerCert;
    while(pCurrentCert != NULL)
    {
        dwVerificationFlags = 0;
        pIssuerCert = CertGetIssuerCertificateFromStore(pServerCert->hCertStore,
                                                        pCurrentCert,
                                                        NULL,
                                                        &dwVerificationFlags);
        if(pIssuerCert == NULL)
        {
            if(pCurrentCert != pServerCert)
            {
                CertFreeCertificateContext(pCurrentCert);
            }
            break;
        }

        if(!CertNameToStr(pIssuerCert->dwCertEncodingType,
                          &pIssuerCert->pCertInfo->Subject,
                          CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
                          szName, sizeof(szName)))
        {
            s_log(LOG_ERR,"**** Error 0x%x building subject name", GetLastError());
        }
        s_log(LOG_NOTICE,"CA subject: %s", szName);
        if(!CertNameToStr(pIssuerCert->dwCertEncodingType,
                          &pIssuerCert->pCertInfo->Issuer,
                          CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
                          szName, sizeof(szName)))
        {
            s_log(LOG_ERR,"**** Error 0x%x building issuer name", GetLastError());
        }
        s_log(LOG_NOTICE,"CA issuer: %s", szName);

        if(pCurrentCert != pServerCert)
        {
            CertFreeCertificateContext(pCurrentCert);
        }
        pCurrentCert = pIssuerCert;
        pIssuerCert = NULL;
    }
} // DisplayCertChain
//-------------------------------------------------------------
 
DWORD
VerifyServerCertificate(
    PCCERT_CONTEXT  pServerCert,
    PSTR            pszServerName,
    DWORD           dwCertFlags)
{
    HTTPSPolicyCallbackData  polHttps;
    CERT_CHAIN_POLICY_PARA   PolicyPara;
    CERT_CHAIN_POLICY_STATUS PolicyStatus;
    CERT_CHAIN_PARA          ChainPara;
    PCCERT_CHAIN_CONTEXT     pChainContext = NULL;

    LPSTR rgszUsages[] = {  szOID_PKIX_KP_SERVER_AUTH,
                            szOID_SERVER_GATED_CRYPTO,
                            szOID_SGC_NETSCAPE };
    DWORD cUsages = sizeof(rgszUsages) / sizeof(LPSTR);
    PWSTR   pwszServerName = NULL;
    DWORD   cchServerName;
    HRESULT   Status;

    if(pServerCert == NULL)
    {
        Status = SEC_E_WRONG_PRINCIPAL;
        goto cleanup;
    }
    if(pszServerName == NULL || strlen(pszServerName) == 0)
    {
        Status = SEC_E_WRONG_PRINCIPAL;
        goto cleanup;
    }
    //server name comes with ":port"
    {
    unsigned int i;
    for(i = 0; i<strlen(pszServerName);i++)
	if(pszServerName[i] == ':')
	    pszServerName[i] = '\0'; 
    }
    cchServerName = MultiByteToWideChar(CP_ACP, 0, pszServerName, -1, NULL, 0);
    pwszServerName = LocalAlloc(LMEM_FIXED, cchServerName * sizeof(WCHAR));
    if(pwszServerName == NULL)
    {
        Status = SEC_E_INSUFFICIENT_MEMORY;
        goto cleanup;
    }
    cchServerName = MultiByteToWideChar(CP_ACP, 0, pszServerName, -1, pwszServerName, cchServerName);
    if(cchServerName == 0)
    {
        Status = SEC_E_WRONG_PRINCIPAL;
        goto cleanup;
    }

    ZeroMemory(&ChainPara, sizeof(ChainPara));
    ChainPara.cbSize = sizeof(ChainPara);
    ChainPara.RequestedUsage.dwType = USAGE_MATCH_TYPE_OR;
    ChainPara.RequestedUsage.Usage.cUsageIdentifier     = cUsages;
    ChainPara.RequestedUsage.Usage.rgpszUsageIdentifier = rgszUsages;

    if(!CertGetCertificateChain(
                            NULL,
                            pServerCert,
                            NULL,
                            pServerCert->hCertStore,
                            &ChainPara,
                            0,
                            NULL,
                            &pChainContext))
    {
        Status = GetLastError();
        s_log(LOG_ERR,"Error 0x%x returned by CertGetCertificateChain!", Status);
        goto cleanup;
    }


    ZeroMemory(&polHttps, sizeof(HTTPSPolicyCallbackData));
    polHttps.cbStruct           = sizeof(HTTPSPolicyCallbackData);
    polHttps.dwAuthType         = AUTHTYPE_SERVER;
    polHttps.fdwChecks          = dwCertFlags;
    polHttps.pwszServerName     = pwszServerName;

    memset(&PolicyPara, 0, sizeof(PolicyPara));
    PolicyPara.cbSize            = sizeof(PolicyPara);
    PolicyPara.pvExtraPolicyPara = &polHttps;

    memset(&PolicyStatus, 0, sizeof(PolicyStatus));
    PolicyStatus.cbSize = sizeof(PolicyStatus);

    if(!CertVerifyCertificateChainPolicy(
                            CERT_CHAIN_POLICY_SSL,
                            pChainContext,
                            &PolicyPara,
                            &PolicyStatus))
    {
        Status = GetLastError();
        s_log(LOG_ERR,"Error 0x%x returned by CertVerifyCertificateChainPolicy!", Status);
        s_log(LOG_ERR,"Server name for checking is %s",pszServerName);
        goto cleanup;
    }

    if(PolicyStatus.dwError){
        Status = PolicyStatus.dwError;
        DisplayWinVerifyTrustError(Status); 
        s_log(LOG_ERR,"Server name for checking is %s",pszServerName);
        goto cleanup;
    }
    Status = SEC_E_OK;
cleanup:

    if(pChainContext)
        CertFreeCertificateChain(pChainContext);
    if(pwszServerName)
        LocalFree(pwszServerName);
    return Status;
} // VerifyServerCertificate


void
DisplayConnectionInfo(
    CtxtHandle *phContext)
{
    SECURITY_STATUS Status;
    SecPkgContext_ConnectionInfo ConnectionInfo;

    Status = g_pSSPI->QueryContextAttributes(phContext,
                                    SECPKG_ATTR_CONNECTION_INFO,
                                    (PVOID)&ConnectionInfo);
    if(Status != SEC_E_OK)
    {
        s_log(LOG_ERR,"Error 0x%x querying connection info", Status);
        return;
    }


    switch(ConnectionInfo.dwProtocol)
    {
        case SP_PROT_TLS1_CLIENT:
            s_log(LOG_NOTICE,"Protocol: TLS1");
            break;

        default:
            s_log(LOG_NOTICE,"Protocol: 0x%x", ConnectionInfo.dwProtocol);
    }

    switch(ConnectionInfo.aiCipher)
    {
        case CALG_G28147: 
            s_log(LOG_NOTICE,"Cipher: Gost 28147-89");
            break;

        default: 
            s_log(LOG_NOTICE,"Cipher: 0x%x", ConnectionInfo.aiCipher);
    }

    s_log(LOG_NOTICE,"Cipher strength: %d", ConnectionInfo.dwCipherStrength);

    switch(ConnectionInfo.aiHash)
    {
        case CALG_GR3411: 
            s_log(LOG_NOTICE,"Hash: Gost R 34.11-94");
            break;

        default: 
            s_log(LOG_NOTICE,"Hash: 0x%x", ConnectionInfo.aiHash);
    }

    s_log(LOG_NOTICE,"Hash strength: %d", ConnectionInfo.dwHashStrength);

    switch(ConnectionInfo.aiExch)
    {
        case CALG_DH_EPHEM:
            s_log(LOG_NOTICE,"Key exchange: DH Ephemeral");
            break;

        default: 
            s_log(LOG_NOTICE,"Key exchange: 0x%x", ConnectionInfo.aiExch);
    }

    s_log(LOG_NOTICE,"Key exchange strength: %d", ConnectionInfo.dwExchStrength);
} //DisplayConnectionInfo



void
GetNewClientCredentials(
    CLI* c,
    CredHandle *phCreds,
    CtxtHandle *phContext)
{
    CredHandle hCreds;
    SecPkgContext_IssuerListInfoEx IssuerListInfo;
    PCCERT_CHAIN_CONTEXT pChainContext;
    CERT_CHAIN_FIND_BY_ISSUER_PARA FindByIssuerPara;
    SCHANNEL_CRED   SchannelCred;
    PCCERT_CONTEXT  pCertContext;
    TimeStamp       tsExpiry;
    SECURITY_STATUS Status;


    Status = g_pSSPI->QueryContextAttributes(phContext,
                                    SECPKG_ATTR_ISSUER_LIST_EX,
                                    (PVOID)&IssuerListInfo);
    if(Status != SEC_E_OK)
    {
        s_log(LOG_ERR,"Error 0x%x querying issuer list info", Status);
        return;
    }


    ZeroMemory(&FindByIssuerPara, sizeof(FindByIssuerPara));

    FindByIssuerPara.cbSize = sizeof(FindByIssuerPara);
    FindByIssuerPara.pszUsageIdentifier = szOID_PKIX_KP_CLIENT_AUTH;
    FindByIssuerPara.dwKeySpec = 0;
    FindByIssuerPara.cIssuer   = IssuerListInfo.cIssuers;
    FindByIssuerPara.rgIssuer  = IssuerListInfo.aIssuers;

    pChainContext = NULL;

    for(;;)
    {
        pChainContext = CertFindChainInStore(c->hMyCertStore,
                                             X509_ASN_ENCODING,
                                             0,
                                             CERT_CHAIN_FIND_BY_ISSUER,
                                             &FindByIssuerPara,
                                             pChainContext);
        if(pChainContext == NULL)
        {
            s_log(LOG_NOTICE,"CertFindChainInStore not find certificate in store. Looking at LOCAL_MACHINE");
            pChainContext = CertFindChainInStore(c->hSysCertStore,
                                                 X509_ASN_ENCODING,
                                                 0,
                                                 CERT_CHAIN_FIND_BY_ISSUER,
                                                 &FindByIssuerPara,
                                                 pChainContext);
        }
        if(pChainContext == NULL)
        {
            s_log(LOG_ERR,"Error 0x%x finding cert chain", GetLastError());
            break;
        }
        s_log(LOG_NOTICE,"certificate chain found");

        pCertContext = pChainContext->rgpChain[0]->rgpElement[0]->pCertContext;
	/* CSA dim:   13  + reserved
	 *   4,    
	 *     ,   
	 */
	memset(&SchannelCred, 0, sizeof(SchannelCred));
        SchannelCred.dwVersion = SCHANNEL_CRED_VERSION;
        SchannelCred.cCreds = 1;
        SchannelCred.paCred = &pCertContext;
	SchannelCred.grbitEnabledProtocols = dwProtocol;
	SchannelCred.dwFlags = SCH_CRED_NO_DEFAULT_CREDS | SCH_CRED_MANUAL_CRED_VALIDATION;

        Status = g_pSSPI->AcquireCredentialsHandle(
                            NULL,                   
                            UNISP_NAME,           
                            SECPKG_CRED_OUTBOUND,   
                            NULL,                   
                            &SchannelCred,          
                            NULL,                   
                            NULL,                   
                            &hCreds,                
                            &tsExpiry);             
        if(Status != SEC_E_OK)
        {
            s_log(LOG_ERR,"**** Error 0x%x returned by AcquireCredentialsHandle", Status);
            continue;
        }
        s_log(LOG_NOTICE,"new schannel credential created");

        g_pSSPI->FreeCredentialsHandle(phCreds);

        *phCreds = hCreds;

        break;
    }
} // End of GetNewClientCredentials

//-------------------------------------------------------------


void
DisplayWinVerifyTrustError(DWORD Status)
{
    LPSTR pszName = NULL;

    switch(Status)
    {
    case CERT_E_EXPIRED:                pszName = "CERT_E_EXPIRED";                 break;
    case CERT_E_VALIDITYPERIODNESTING:  pszName = "CERT_E_VALIDITYPERIODNESTING";   break;
    case CERT_E_ROLE:                   pszName = "CERT_E_ROLE";                    break;
    case CERT_E_PATHLENCONST:           pszName = "CERT_E_PATHLENCONST";            break;
    case CERT_E_CRITICAL:               pszName = "CERT_E_CRITICAL";                break;
    case CERT_E_PURPOSE:                pszName = "CERT_E_PURPOSE";                 break;
    case CERT_E_ISSUERCHAINING:         pszName = "CERT_E_ISSUERCHAINING";          break;
    case CERT_E_MALFORMED:              pszName = "CERT_E_MALFORMED";               break;
    case CERT_E_UNTRUSTEDROOT:          pszName = "CERT_E_UNTRUSTEDROOT";           break;
    case CERT_E_CHAINING:               pszName = "CERT_E_CHAINING";                break;
    case TRUST_E_FAIL:                  pszName = "TRUST_E_FAIL";                   break;
    case CERT_E_REVOKED:                pszName = "CERT_E_REVOKED";                 break;
    case CERT_E_UNTRUSTEDTESTROOT:      pszName = "CERT_E_UNTRUSTEDTESTROOT";       break;
    case CERT_E_REVOCATION_FAILURE:     pszName = "CERT_E_REVOCATION_FAILURE";      break;
    case CERT_E_CN_NO_MATCH:            pszName = "CERT_E_CN_NO_MATCH";             break;
    case CERT_E_WRONG_USAGE:            pszName = "CERT_E_WRONG_USAGE";             break;
    default:                            pszName = "(unknown)";                      break;
    }

    s_log(LOG_ERR,"Error 0x%x (%s) returned by CertVerifyCertificateChainPolicy!", 
        Status, pszName);
} //DisplayWinVerifyTrustError

void print_state(CLI* c){
    s_log(LOG_NOTICE,"-----------Print_State_start----------------------------");    
    s_log(LOG_NOTICE,"sock_rd rd=%d wr=%d",c->sock_rfd->rd,c->sock_rfd->wr);    
    s_log(LOG_NOTICE,"sock_wr rd=%d wr=%d",c->sock_wfd->rd,c->sock_wfd->wr);    
    s_log(LOG_NOTICE,"ssl_rd rd=%d wr=%d",c->ssl_rfd->rd,c->ssl_rfd->wr);    
    s_log(LOG_NOTICE,"ssl_wr rd=%d wr=%d",c->ssl_wfd->rd,c->ssl_wfd->wr);    
    s_log(LOG_NOTICE,"ssl_ptr =%d",c->ssl_ptr);    
    s_log(LOG_NOTICE,"socket_ptr r=%d",c->sock_ptr);    
    s_log(LOG_NOTICE,"sock_bytes=%d ssl_bytes= %d",c->sock_bytes,c->ssl_bytes);    
    s_log(LOG_NOTICE,"-----------Print_State_finish----------------------------");    
}


int SSPI_shutdown(CLI* c){
    DWORD           dwType;
    PCHAR           pbMessage;
    DWORD           cbMessage;
    int		    cbData =0;

    SecBufferDesc   OutBuffer;
    SecBuffer       OutBuffers[1];
    DWORD           dwSSPIFlags;
    unsigned long   dwSSPIOutFlags;
    TimeStamp       tsExpiry;
    DWORD           Status;
#ifdef HSM_SERVER
    uid_t saved_user = 0;
    gid_t saved_group = 0;
#endif

    dwType = SCHANNEL_SHUTDOWN;

    OutBuffers[0].pvBuffer   = &dwType;
    OutBuffers[0].BufferType = SECBUFFER_TOKEN;
    OutBuffers[0].cbBuffer   = sizeof(dwType);

    OutBuffer.cBuffers  = 1;
    OutBuffer.pBuffers  = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

    Status = g_pSSPI->ApplyControlToken(&(c->hContext), &OutBuffer);

    if(FAILED(Status)){ 
        s_log(LOG_ERR,"**** Error 0x%x returned by ApplyControlToken", Status);
        return 0;
    }


    dwSSPIFlags =   ASC_REQ_SEQUENCE_DETECT     |
                    ASC_REQ_REPLAY_DETECT       |
                    ASC_REQ_CONFIDENTIALITY     |
                    ASC_REQ_EXTENDED_ERROR      |
                    ASC_REQ_ALLOCATE_MEMORY     |
                    ASC_REQ_STREAM;

    OutBuffers[0].pvBuffer   = NULL;
    OutBuffers[0].BufferType = SECBUFFER_TOKEN;
    OutBuffers[0].cbBuffer   = 0;

    OutBuffer.cBuffers  = 1;
    OutBuffer.pBuffers  = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;
#ifdef HSM_SERVER
    if(c->opt->setegid_group && c->opt->seteuid_user){
        saved_user = geteuid ();
        saved_group = getegid ();
        seteuid(0);
        setegid(c->opt->setegid_group); 
        seteuid(c->opt->seteuid_user); 
    }
#endif
    if(c->opt->option.client){
    Status = g_pSSPI->AcceptSecurityContext(
                    &(c->hClientCreds),
                    &(c->hContext),
                    NULL,
                    dwSSPIFlags,
                    SECURITY_NATIVE_DREP,
                    NULL,
                    &OutBuffer,
                    &dwSSPIOutFlags,
                    &tsExpiry);
    }else{
     Status = g_pSSPI->AcceptSecurityContext(
                    &(c->opt->hServerCreds),
                    &(c->hContext),
                    NULL,
                    dwSSPIFlags,
                    SECURITY_NATIVE_DREP,
                    NULL,
                    &OutBuffer,
                    &dwSSPIOutFlags,
                    &tsExpiry);

    }
#ifdef HSM_SERVER
    if(c->opt->setegid_group && c->opt->seteuid_user){
        seteuid(0);
        setegid(saved_group); 
        seteuid(saved_user); 
    }
#endif
    if(FAILED(Status)){
        s_log(LOG_ERR,"**** Error 0x%x returned by AcceptSecurityContext", Status);
        return 0;
    }

    pbMessage = OutBuffers[0].pvBuffer;
    cbMessage = OutBuffers[0].cbBuffer;


    if(pbMessage != NULL && cbMessage != 0){
        cbData = send(c->ssl_wfd->fd,pbMessage,cbMessage,0);
        if(IS_SOCKET_ERROR(cbData)|| cbData == 0){
            Status = WSAGetLastError();
            s_log(LOG_ERR,"**** Error %d sending close notify", Status);
            g_pSSPI->FreeContextBuffer(pbMessage);
            return 0;
        }
        s_log(LOG_NOTICE,"%d bytes of close_notify data sent", cbData);
        g_pSSPI->FreeContextBuffer(pbMessage);
    }
    return 1;
} //SSPI_shutdown 


int SSPI_write_ex(CLI* c){
    SecBufferDesc   Message;
    SecBuffer       Buffers[4];
    SECURITY_STATUS scRet;
    int             err = 0;
    s_log(LOG_NOTICE,"SSPI_write start");
    if(c->opt->option.without_enc){
        err = send( c->ssl_wfd->fd,
                c->sock_buff,
                c->sock_ptr, 0 );
        switch(err){
        case -1:
            if(errno == EAGAIN){
            s_log(LOG_ERR,"ssl_write will block socket -> return");
            c->sspi_want_wr = 1;
            return 0;
            break;
            }
            s_log(LOG_ERR, "socket error = %d  when send data to ssl socket",errno);
            return (-1);
        case 0:
            s_log(LOG_ERR,"Zero byte send");
            c->sspi_want_wr = 1;
            return 0;
        default:
            s_log(LOG_NOTICE,"actualy send %d bytes to ssl socket without_enc", err);
            if(err < c->sock_ptr){//the send data was less then message
            c->sspi_want_wr = 1;
            return err;
            }
            if (c->sock_ptr == 0){ // check for new data for encrypt
            c->sspi_want_wr = 0;
            return err;
            }
            break;
        }
        return err;
    }//without_enc


    if(c->sspi_want_wr){//there is data to send from last call
	err = send( c->ssl_wfd->fd,
		    c->wIoBuffer,
		    c->cbwIoBuffer, 0 );
	switch(err){
	    case -1:
		if(errno == EAGAIN){
		s_log(LOG_ERR,"ssl_write will block socket -> return");
		c->sspi_want_wr = 1;
		return 0;
		break;
		}
	    s_log(LOG_ERR, "socket error = %d  when send encrypt data",errno);
	    return (-1);
	case 0:
	    s_log(LOG_ERR,"Zero byte send");
	    c->sspi_want_wr = 1;
	    return 0;
	default:
	    s_log(LOG_NOTICE,"actualy send %d bytes to ssl socket", err);
	    if(err < c->cbwIoBuffer){//the send data was less then message
		c->sspi_want_wr = 1;
		memmove(c->wIoBuffer,c->wIoBuffer+err,c->cbwIoBuffer-err);
		c->cbwIoBuffer-=err;
		return 0;
		}
	    if (c->sock_ptr == 0){ // check for new data for encrypt
		c->sspi_want_wr = 0;
		return 0;
		}
	    break;
	}
    }
    memset(c->writeBuffer, 0, Sizes.cbHeader);

    Message.ulVersion = SECBUFFER_VERSION;
    Message.cBuffers = 4;
    Message.pBuffers = Buffers;
    
    Buffers[0].pvBuffer = c->writeBuffer;
    Buffers[0].cbBuffer = Sizes.cbHeader;
    Buffers[0].BufferType = SECBUFFER_STREAM_HEADER;
    memcpy((c->writeBuffer+Sizes.cbHeader),c->sock_buff,c->sock_ptr);
    s_log(LOG_DEBUG,"SSPI_write data  is %.4s",(c->writeBuffer+Sizes.cbHeader));
    Buffers[1].pvBuffer = c->writeBuffer + Sizes.cbHeader;
    Buffers[1].cbBuffer = c->sock_ptr;
    Buffers[1].BufferType = SECBUFFER_DATA;

    Buffers[2].pvBuffer = c->writeBuffer + Sizes.cbHeader + c->sock_ptr;
    Buffers[2].cbBuffer = Sizes.cbTrailer;
    Buffers[2].BufferType = SECBUFFER_STREAM_TRAILER;

    Buffers[3].BufferType = SECBUFFER_EMPTY;

    scRet = g_pSSPI->EncryptMessage(&(c->hContext),
                                    0,
                                    &Message,
                                    0);
    if (FAILED(scRet)){
        s_log(LOG_ERR," EncryptMessage failed with 0x%x", scRet );
        return -1;
    }
    //send new encrypted data
    err = send( c->ssl_wfd->fd,
		    c->writeBuffer,
		    Buffers[0].cbBuffer+Buffers[1].cbBuffer+ Buffers[2].cbBuffer, 0 );
    switch(err){
    case -1:
        if(errno == EAGAIN){
                s_log(LOG_ERR,"SSPI_write: socket will block -> return");
		c->cbwIoBuffer = Buffers[0].cbBuffer+Buffers[1].cbBuffer + Buffers[2].cbBuffer;//calculate lenght of encrypted message
		memcpy(c->wIoBuffer,c->writeBuffer,c->cbwIoBuffer);
		c->sspi_want_wr = 1;
                return c->sock_ptr;
                }
        s_log(LOG_ERR, "socket error = %d  when send encrypt data",errno);
        return (-1);
    case 0:
        s_log(LOG_DEBUG,"Zero byte send -> retry");
	c->cbwIoBuffer = Buffers[0].cbBuffer+Buffers[1].cbBuffer + Buffers[2].cbBuffer;//calculate lenght of encrypted message
	memcpy(c->wIoBuffer,c->writeBuffer,c->cbwIoBuffer);
	c->sspi_want_wr = 1;
	return c->sock_ptr;
    default:
        if(err>=(c->cbwIoBuffer = Buffers[0].cbBuffer+Buffers[1].cbBuffer + Buffers[2].cbBuffer)){
	   c->sspi_want_wr = 0;
	   s_log(LOG_DEBUG,"send all data after encrypt");
	   return c->sock_ptr;
	   }
	memcpy(c->wIoBuffer,c->writeBuffer+err,c->cbwIoBuffer-err);
	c->sspi_want_wr = 1;
	c->cbwIoBuffer-=err;
	s_log(LOG_DEBUG,"actualy send %d byte of encypt data", err);
        return (c->sock_ptr);
       }
    
}// end of SSPI_write

int SSPI_write (CLI* c){
    int num;
#ifdef HSM_SERVER
    uid_t saved_user = 0;
    gid_t saved_group = 0;

    if(c->opt->setegid_group && c->opt->seteuid_user){
        saved_user = geteuid ();
        saved_group = getegid ();
        seteuid(0);
        setegid(c->opt->setegid_group); 
        seteuid(c->opt->seteuid_user); 
    }
#endif
    num = SSPI_write_ex (c);
#ifdef HSM_SERVER
    if(c->opt->setegid_group && c->opt->seteuid_user){
        seteuid(0);
        setegid(saved_group); 
        seteuid(saved_user); 
    }
#endif
    return num;
}


int SSPI_read_ex(CLI* c) {
    SecBufferDesc   Message;
    SecBuffer       Buffers[4];
    SECURITY_STATUS scRet;
    int err;
    Message.ulVersion = SECBUFFER_VERSION;
    Message.cBuffers = 4;
    Message.pBuffers = Buffers;
    c->cbrIoBuffer = 0;
    s_log(LOG_DEBUG,"SSPI_read start");
    if(c->opt->option.without_enc){
        err = recv(c->ssl_rfd->fd, c->ssl_buff + c->ssl_ptr, (2*BUFFSIZE - c->ssl_ptr), 0);
        s_log(LOG_DEBUG,"recv ok on SSPI_read err= %d",err);
        if (IS_SOCKET_ERROR(err)) 
        {
            if(errno == EAGAIN){
            return 0;
	    }
        s_log(LOG_ERR,"recv failed: %d errno =%x ", err, errno);
        return -1;
        }
        if(err == 0){//connection close but there is not complete data for decrypt -> error
            s_log(LOG_ERR,"recv return 0 but there is not complete data for decrypt");
            return -3;
        }
        if (err>0){
            s_log(LOG_NOTICE,"Received %d  bytes from ssl socket without_enc", err); 
        }    
        return err;
    }//without_enc mode finish

    //add data from last call
    if (c->ExLen >0){
        memcpy(c->rIoBuffer, c->ExDataBuffer,c->ExLen);
        s_log(LOG_DEBUG,"add data from  last call = %d", c->ExLen);
        c->cbrIoBuffer = c->ExLen;
        c->ExLen = 0;
        }
    scRet = SEC_E_INCOMPLETE_MESSAGE;
    // initialize buffer 
    Buffers[0].pvBuffer = c->rIoBuffer;
    Buffers[0].cbBuffer = c->cbrIoBuffer;
    Buffers[0].BufferType = SECBUFFER_DATA;

    Buffers[1].BufferType = SECBUFFER_EMPTY;
    Buffers[2].BufferType = SECBUFFER_EMPTY;
    Buffers[3].BufferType = SECBUFFER_EMPTY;

    if ((c->cbrIoBuffer > 0 && c->sspi_buff_st)|| c->ExData) //not call decrypt if we know that data is incomplete
	scRet = g_pSSPI->DecryptMessage(&c->hContext, &Message, 0, NULL);
    c->ExData = 0;// reset the flag from handshake
    if(scRet == SEC_E_OK){	
	if(Buffers[3].BufferType == SECBUFFER_EXTRA){
	    s_log(LOG_DEBUG,"Recieve %d bytes EXTRA_DATA",Buffers[3].cbBuffer);
	    c->extra_data++;
	    c->sspi_want_rd=1;//put the flag for want_read
	    memcpy((c->ssl_buff+c->ssl_ptr),Buffers[1].pvBuffer,Buffers[1].cbBuffer);
	    memcpy(c->ExDataBuffer,(c->rIoBuffer + (c->cbrIoBuffer - Buffers[3].cbBuffer)), Buffers[3].cbBuffer);
	    c->ExLen = Buffers[3].cbBuffer;
	    c->sspi_buff_st = 1;
	    return (Buffers[1].cbBuffer);
	}
	else {
	    // Technically, walk through the buffer desc, looking for the first
	    // buffer labelled SECBUFFER_DATA.  Practically, it will be the second
	    // buffer section.
	    memcpy((c->ssl_buff+c->ssl_ptr),Buffers[1].pvBuffer,Buffers[1].cbBuffer);
	    s_log(LOG_DEBUG,"SSPI_read data in ssl_buff is %.4s",(c->ssl_buff+c->ssl_ptr));
	    c->sspi_want_rd=0;//put the flag for want_read: all data decrypt
	    c->sspi_buff_st = 0;
	    return (Buffers[1].cbBuffer);
	}
    }
    if(scRet == SEC_E_INCOMPLETE_MESSAGE)
    {
	c->incomp_mess++;
	// try to get new data
	err = recv(c->ssl_rfd->fd, c->rIoBuffer + c->cbrIoBuffer, (2*BUFFSIZE+2*Sizes.cbHeader+2*Sizes.cbTrailer-c->cbrIoBuffer-c->ssl_ptr), 0);
	s_log(LOG_DEBUG,"recv ok on SSPI_read err= %d",err);
	if (IS_SOCKET_ERROR(err)) 
	{
	    if(errno == EAGAIN){
	      c->sspi_want_rd = 1;
	      c->sspi_buff_st=0;
	      memcpy(c->ExDataBuffer,c->rIoBuffer, c->cbrIoBuffer);
	      c->ExLen = c->cbrIoBuffer;
	      return 0;
	    }
            s_log(LOG_ERR,"recv failed: %d errno =%x ", err, errno);
#ifdef _WIN32
          if(errno ==0 && c->cbrIoBuffer){
	    c->sspi_want_rd = 1;
	    c->sspi_buff_st=0;
	    memcpy(c->ExDataBuffer,c->rIoBuffer, c->cbrIoBuffer);
	    c->ExLen = c->cbrIoBuffer;
            s_log(LOG_DEBUG,"Win issue with nonblocking socket, lets try wait and read again");
	    return 0;
          }
#endif 
	  return -1;
	}
	if(err == 0){//connection close but there is not complete data for decrypt
        if(!(2*BUFFSIZE+2*Sizes.cbHeader+2*Sizes.cbTrailer-c->cbrIoBuffer-c->ssl_ptr)){
            s_log(LOG_NOTICE,"recv return 0 and ask %d",(int)(2*BUFFSIZE+2*Sizes.cbHeader+2*Sizes.cbTrailer-c->cbrIoBuffer-c->ssl_ptr));
            if(c->cbrIoBuffer){
                memcpy(c->ExDataBuffer,c->rIoBuffer,c->cbrIoBuffer);
                c->ExLen = c->cbrIoBuffer;
                c->sspi_want_rd = 1;
            }
            return 0;
        }
	    s_log(LOG_ERR,"recv return 0 and ask more but there is not complete data for decrypt");
	    return -3;
	    }
	if (err>0){
	    s_log(LOG_NOTICE,"Received %d  bytes from ssl socket", err);        
	    c->cbrIoBuffer += err;
	    Buffers[0].pvBuffer = c->rIoBuffer;
	    Buffers[0].cbBuffer = c->cbrIoBuffer;
	    Buffers[0].BufferType = SECBUFFER_DATA;

	    Buffers[1].BufferType = SECBUFFER_EMPTY;
	    Buffers[2].BufferType = SECBUFFER_EMPTY;
	    Buffers[3].BufferType = SECBUFFER_EMPTY;
		       
	    scRet = g_pSSPI->DecryptMessage(&c->hContext, &Message, 0, NULL);
	    if(scRet == SEC_E_OK){	
		if(Buffers[3].BufferType == SECBUFFER_EXTRA){
		    s_log(LOG_DEBUG,"Recieve %d bytes EXTRA_DATA",Buffers[3].cbBuffer);
		    c->extra_data++;
		    memcpy((c->ssl_buff+c->ssl_ptr),Buffers[1].pvBuffer,Buffers[1].cbBuffer);
		    memcpy(c->ExDataBuffer,(c->rIoBuffer + (c->cbrIoBuffer - Buffers[3].cbBuffer)), Buffers[3].cbBuffer);
		    c->ExLen = Buffers[3].cbBuffer;
		    s_log(LOG_DEBUG,"data read return  %d, c->ssl_ptr = %d",Buffers[1].cbBuffer,c->ssl_ptr);
		    c->sspi_want_rd=1;//put the flag for want_read
		    c->sspi_buff_st = 1;
		    return (Buffers[1].cbBuffer);
	    }
	    else {
		    // Technically, walk through the buffer desc, looking for the first
		    // buffer labelled SECBUFFER_DATA.  Practically, it will be the second
		    // buffer section.
		    memcpy((c->ssl_buff+c->ssl_ptr),Buffers[1].pvBuffer,Buffers[1].cbBuffer);
		    s_log(LOG_DEBUG,"SSPI_read data in ssl_buff is %.4s",(c->ssl_buff+c->ssl_ptr));
		    c->sspi_want_rd=0;//put the flag for want_read: all data decrypt
		    c->sspi_buff_st = 0;
		    return (Buffers[1].cbBuffer);
		}
	    }
	    if(scRet == SEC_E_INCOMPLETE_MESSAGE){
		c->sspi_want_rd = 1;
		memcpy(c->ExDataBuffer,c->rIoBuffer, c->cbrIoBuffer);
		c->incomp_mess++;
		c->ExLen = c->cbrIoBuffer;
		c->sspi_buff_st = 0; //set flag for next call for incomplete data
		return 0;
		}

	}
    } 
    if(scRet == SEC_I_CONTEXT_EXPIRED){
        s_log(LOG_ERR, "SEC_I_CONTEXT_EXPIRED,");
        c->sspi_want_rd = 0;
        c->sspi_buff_st = 0;
        return -2;
    }
    // it will depend on implementation of CSP
    if( scRet == SEC_I_RENEGOTIATE){
	SecBuffer  ExtraData;
        s_log(LOG_NOTICE,"Client request RENEGOTIATE");        
        if(ClientHandshakeLoop(c->remote_fd.fd,
                          &(c->hClientCreds),
			  &(c->hContext),
			  FALSE,
			  &ExtraData,
                              c))
                {
                s_log(LOG_ERR,"Couldn't complete RENEGOTIATE");
                return -1;
                }
       return 0; 
    }    
    if(scRet != SEC_E_OK){
        s_log(LOG_ERR,"Couldn't decrypt, error %x", scRet);
        return -1;
    }
    return -1; // to disable warning of reaching end of non void function
} //end of SSPI_read


int SSPI_read(CLI* c) {
    int num;
#ifdef HSM_SERVER
    uid_t saved_user = 0;
    gid_t saved_group = 0;

    if(c->opt->setegid_group && c->opt->seteuid_user){
        saved_user = geteuid ();
        saved_group = getegid ();
        seteuid(0);
        setegid(c->opt->setegid_group); 
        seteuid(c->opt->seteuid_user); 
    }
#endif
    num = SSPI_read_ex (c);

#ifdef HSM_SERVER
    if(c->opt->setegid_group && c->opt->seteuid_user){
        seteuid(0);
        setegid(saved_group); 
        seteuid(saved_user); 
    }
#endif
    return num;
}

void handshake (CLI* c){

    SecBuffer  ExtraData;
    SECURITY_STATUS Status;
    
    PCCERT_CONTEXT pRemoteCertContext = NULL;

    s_log(LOG_DEBUG, "start SSPI connect");
    if(CreateClientCredentials(c)){
        s_log(LOG_ERR, "Error creating credentials");
        longjmp(c->err,1);
    }
    if(PerformClientHandshake(c->remote_fd.fd,
                              &(c->hClientCreds),
                              c->opt->remote_address,
                              &(c->hContext),
                              &ExtraData,
				c))
    {
        s_log(LOG_ERR,"Error performing handshake");
        longjmp(c->err,1);
    }
    s_log(LOG_NOTICE, "PerformClientHandshake finish ");
    if(c->opt->verify_level == SSL_VERIFY_NONE){
        s_log(LOG_NOTICE, "Verify_level = 0, skipping Server certificate verification");
        return;
    }

    Status = g_pSSPI->QueryContextAttributes(&(c->hContext),
                                             SECPKG_ATTR_REMOTE_CERT_CONTEXT,
                                             (PVOID)&pRemoteCertContext);
    if(Status != SEC_E_OK){
        s_log(LOG_ERR,"Error 0x%x querying remote certificate", Status);
        longjmp(c->err,1);
    }

    DisplayCertChain(pRemoteCertContext, FALSE);
    Status = VerifyServerCertificate(pRemoteCertContext,
                              c->opt->remote_address,
                                     0);
    if(Status){
        s_log(LOG_ERR,"**** Error 0x%x authenticating server credentials!", Status);
        CertFreeCertificateContext(pRemoteCertContext);
        longjmp(c->err,1);
    }

    CertFreeCertificateContext(pRemoteCertContext);
    pRemoteCertContext = NULL;
    DisplayConnectionInfo(&(c->hContext));
    s_log(LOG_DEBUG,"Handshake_done");

}//Handhsake

void accept_handshake(CLI* c){
//    uid_t saved_user = 0;
//    gid_t saved_group = 0;
    BOOL res;

    s_log(LOG_DEBUG,"accept_handshake start");
#ifdef HSM_SERVER
    if(c->opt->setegid_group && c->opt->seteuid_user){
//        saved_user = geteuid ();
//        saved_group = getegid ();
        seteuid(0);
        setegid(c->opt->setegid_group); 
        seteuid(c->opt->seteuid_user); 
    }
#endif
    res = SSPINegotiateLoop ((c->local_rfd.fd),
			  &(c->hContext),
			  &(c->opt->hServerCreds),
			  TRUE,
			  TRUE,
			  c);
//We use pam inside SSPINegotiateLoop so we under good user atm
#if 0  
    if(c->opt->setegid_group && c->opt->seteuid_user){
        seteuid(0);
        setegid(saved_group); 
        seteuid(saved_user); 
    }
#endif

    if (!res)
    {
	s_log(LOG_ERR,"Couldn't connect");
        longjmp(c->err,1);
    }
    s_log(LOG_NOTICE,"User validation finish ExLen = %d",c->ExLen);
} //End of accept_handshake


BOOL
SSPINegotiateLoop(
    SOCKET          Socket,
    PCtxtHandle     phContext,
    PCredHandle     phCred,
    BOOL            fDoInitialRead,
    BOOL            NewContext,
    CLI* c)
{
    TimeStamp            tsExpiry;
    SECURITY_STATUS      scRet = SEC_I_CONTINUE_NEEDED;
    SecBufferDesc        InBuffer;
    SecBufferDesc        OutBuffer;
    SecBuffer            InBuffers[2];
    SecBuffer            OutBuffers[1];
    int                err = 0;
    int                  watchdog = 0;
    int                  wRes; //For s_poll_wait
    s_poll_set           fds; /* File descriptors */
    BOOL                 fDoRead;
    BOOL                 fInitContext = NewContext;

    DWORD                dwSSPIFlags;
    unsigned long        dwSSPIOutFlags;
    
    c->cbwIoBuffer = 0;//Lenght of recv message
    s_log(LOG_DEBUG,"SSPINegotiate start");
    fDoRead = fDoInitialRead;

    dwSSPIFlags =   ASC_REQ_SEQUENCE_DETECT        |
                    ASC_REQ_REPLAY_DETECT      |
                    ASC_REQ_CONFIDENTIALITY  |
                    ASC_REQ_EXTENDED_ERROR    |
                    ASC_REQ_ALLOCATE_MEMORY  |
                    ASC_REQ_STREAM ;
    if(c->opt->verify_level != SSL_VERIFY_NONE)
	dwSSPIFlags |= ASC_REQ_MUTUAL_AUTH;

    OutBuffer.cBuffers = 1;
    OutBuffer.pBuffers = OutBuffers;
    OutBuffer.ulVersion = SECBUFFER_VERSION;

    s_poll_zero(&fds);
    s_poll_add(&fds,Socket,1,0);
    while( scRet == SEC_I_CONTINUE_NEEDED ||
            scRet == SEC_E_INCOMPLETE_MESSAGE ||
            scRet == SEC_I_INCOMPLETE_CREDENTIALS) 
    {

        if(0 == c->cbwIoBuffer || scRet == SEC_E_INCOMPLETE_MESSAGE)
        {
            if(fDoRead)
            { 
                wRes = s_poll_wait(&fds,200,0);
                switch(wRes) {
                    case -1:
                        sockerror("hanshake_loop: s_poll_wait error");
                        return FALSE;
                    case 0: /* timeout */
                        s_log(LOG_INFO, "handshake loop s_poll_wait timeout: connection reset");
                        return FALSE;
                }
                err = recv(Socket, c->wIoBuffer+c->cbwIoBuffer, BUFFSIZE - c->cbwIoBuffer, 0);
                s_log(LOG_DEBUG,"reading in SSPINeg err = %d",err);

                if (err<0){ 
                    s_log(LOG_ERR," recv failed: %x err = %ux", err, GetLastError() );
                    return FALSE;
                }
                if(err>0){
                    s_log(LOG_DEBUG," Recieve %d bytes from client on SSPINegotiateLoop",err);
                    //PrintHexDump(err,IoBuffer+cbIoBuffer);
                    c->cbwIoBuffer += err;
                    watchdog =0;
                }
                if(++watchdog> 10){
                    s_log(LOG_ERR,"wathcdog limit finish on SSPINegotiateLoop");
                    return FALSE;
                }
            }
            else
                fDoRead = TRUE;
        }
        
        InBuffers[0].pvBuffer = c->wIoBuffer;
        InBuffers[0].cbBuffer = c->cbwIoBuffer;
        InBuffers[0].BufferType = SECBUFFER_TOKEN;

        InBuffers[1].pvBuffer   = NULL;
        InBuffers[1].cbBuffer   = 0;
        InBuffers[1].BufferType = SECBUFFER_EMPTY;

        InBuffer.cBuffers        = 2;
        InBuffer.pBuffers        = InBuffers;
        InBuffer.ulVersion       = SECBUFFER_VERSION;

        OutBuffers[0].pvBuffer   = NULL;
        OutBuffers[0].BufferType = SECBUFFER_TOKEN;
        OutBuffers[0].cbBuffer   = 0;

        scRet = g_pSSPI->AcceptSecurityContext(
                        phCred,
                        (fInitContext?NULL:phContext),
                        &InBuffer,
                        dwSSPIFlags,
                        SECURITY_NATIVE_DREP,
                        (fInitContext?phContext:NULL),
                        &OutBuffer,
                        &dwSSPIOutFlags,
                        &tsExpiry);
        s_log(LOG_DEBUG,"AcceptSecurityContext finish, scRet = %d",scRet);
        if(scRet == SEC_E_INCOMPLETE_MESSAGE){
#ifdef USE_WIN32
	    Sleep(200);
#else
            usleep(100000);
#endif
            s_log(LOG_DEBUG,"No data sleep, incomp_mess=%d",c->incomp_mess);
            c->incomp_mess++;
        }
        if (scRet != SEC_E_INCOMPLETE_MESSAGE)
            fInitContext = FALSE;

        if ( scRet == SEC_E_OK ||
             scRet == SEC_I_CONTINUE_NEEDED ||
             (FAILED(scRet) && (0 != (dwSSPIOutFlags & ISC_RET_EXTENDED_ERROR))))
        {
            //Last step of handshake need verify remote cert
            //function used longjump on error so no return code
            if(scRet == SEC_E_OK){
                verify_remote_cert(c);
            }
            if  (OutBuffers[0].cbBuffer != 0    &&
                 OutBuffers[0].pvBuffer != NULL )
            {
                err = send( Socket,
                            OutBuffers[0].pvBuffer,
                            OutBuffers[0].cbBuffer,
                            0 );

                s_log(LOG_NOTICE,"Send %d handshake bytes to client", OutBuffers[0].cbBuffer);
                //PrintHexDump(err,OutBuffers[0].pvBuffer);
              
                g_pSSPI->FreeContextBuffer( OutBuffers[0].pvBuffer );
                OutBuffers[0].pvBuffer = NULL;
            }
        }
        if ( scRet == SEC_E_OK )
        {
            if ( InBuffers[1].BufferType == SECBUFFER_EXTRA )
            {
                    c->extra_data++;
                    s_log(LOG_DEBUG,"recv extra date after handshake");
                    memcpy(c->ExDataBuffer,
                           (LPBYTE)(c->wIoBuffer + (c->cbwIoBuffer - InBuffers[1].cbBuffer)),
                            InBuffers[1].cbBuffer);
                    c->ExLen = InBuffers[1].cbBuffer;
                    c->ExData = 1;//flag for skip poll in transfer in first loop
                    c->sspi_buff_st = 1;
                    s_log(LOG_NOTICE,"%d bytes of app data was bundled with handshake data",c->ExLen);
                    
            }
            else
            {
                c->cbwIoBuffer = 0;
            }
            return TRUE;
        }
        else if (FAILED(scRet) && (scRet != SEC_E_INCOMPLETE_MESSAGE))
        {
            s_log(LOG_ERR,"Accept Security Context Failed with error code %x", scRet);
            return FALSE;
        }
        if ( scRet != SEC_E_INCOMPLETE_MESSAGE &&
             scRet != SEC_I_INCOMPLETE_CREDENTIALS)
        {


            if ( InBuffers[1].BufferType == SECBUFFER_EXTRA )
            {
                c->extra_data++;
                memcpy(c->wIoBuffer,
                       (LPBYTE) (c->wIoBuffer + (c->cbwIoBuffer - InBuffers[1].cbBuffer)),
                        InBuffers[1].cbBuffer);
                c->cbwIoBuffer = InBuffers[1].cbBuffer;
            }
            else
            {

                c->cbwIoBuffer = 0;
            }
        }
    }
    return FALSE;
} //SSPINegotiateLoop



//-------------------------------------------------------------
BOOL
LoadSecurityLibrary(void)
{
    INIT_SECURITY_INTERFACE         pInitSecurityInterface;
#ifdef USE_WIN32
    WCHAR lpszDLL[MAX_PATH];
    wcscpy (lpszDLL, NT4_DLL_NAME );
    g_hSecurity = LoadLibrary(lpszDLL);
    if(g_hSecurity == NULL)
    {
       s_log(LOG_ERR,"Error 0x%x loading %s.", GetLastError(), lpszDLL);
        return FALSE;
    }

    pInitSecurityInterface = (INIT_SECURITY_INTERFACE)GetProcAddress(
                                    g_hSecurity,
                                    "InitSecurityInterfaceW");
    
#else
    pInitSecurityInterface=InitSecurityInterfaceA;
#endif

    if(pInitSecurityInterface == NULL)
    {
        s_log(LOG_ERR,"Error 0x%x reading InitSecurityInterface entry point.", 
               GetLastError());
        return FALSE;
    }

    g_pSSPI = pInitSecurityInterface();

    if(g_pSSPI == NULL)
    {
        s_log(LOG_ERR,"Error 0x%x reading security interface.",
               GetLastError());
        return FALSE;
    }

    return TRUE;
}//LoadSecurityLibrary
//-------------------------------------------------------------

void
UnloadSecurityLibrary(void)
{
#ifdef USE_WIN32
    FreeLibrary(g_hSecurity);
    g_hSecurity = NULL;
#endif
} //UnloadSecurityLibrary

BOOL validate_connect(CLI* c){
    PCCERT_CONTEXT pCert = NULL;
    SECURITY_STATUS Status;

    if(c->opt->verify_level == SSL_VERIFY_NONE){
        s_log(LOG_NOTICE, "Verify_level = 0, skipping client certificate verification");
        return TRUE;
    }

    Status = g_pSSPI->QueryContextAttributes(&c->hContext,
                                             SECPKG_ATTR_REMOTE_CERT_CONTEXT,
                                             (PVOID)&pCert);
    if(Status != SEC_E_OK)
    {
        if(c->opt->verify_level & SSL_VERIFY_FAIL_IF_NO_PEER_CERT){
            s_log(LOG_ERR,"Error 0x%x querying remote certificate", Status);
            CertFreeCertificateContext(pCert);
            return FALSE;
        }
        return TRUE;
    }
#ifdef USE_PAM
    {
    char* env;
    LPTSTR cert;
    DWORD cbCert;
    if(!CryptBinaryToString(pCert->pbCertEncoded,pCert->cbCertEncoded,CRYPT_STRING_BASE64,NULL,&cbCert)){ 
         CertFreeCertificateContext(pCert);
         s_log(LOG_ERR,"CryptBinaryToString error");
         return FALSE;
    }
    cert = malloc(cbCert);
    if (!cert){
         CertFreeCertificateContext(pCert);
         s_log(LOG_ERR,"malloc failed");
         return FALSE;
     } 
    if(!CryptBinaryToString(pCert->pbCertEncoded,pCert->cbCertEncoded,CRYPT_STRING_BASE64,cert,&cbCert)){
         CertFreeCertificateContext(pCert);
         s_log(LOG_ERR,"CryptBinaryToString error");
         free(cert);
         return FALSE;
     }
    env = malloc(cbCert+sizeof("SSL_CLIENT_CERT="));
    if(!env){
         CertFreeCertificateContext(pCert);
         s_log(LOG_ERR,"malloc failed");
         free(cert);
         return FALSE;
    }
    snprintf(env,cbCert+17,"SSL_CLIENT_CERT=%s",cert);

    if(pam_putenv(c->pamh,env) != 0)
        s_log(LOG_ERR,"setenv err = %d",errno);
    free(env);
    free(cert);
    }
#endif
   
    if(SEC_E_OK != ssl_verify_by_level(c, pCert)){
        CertFreeCertificateContext(pCert);
        return FALSE;
    }
#ifdef HSM_SERVER
    {
	DWORD i;
	DWORD cbUsage;
	int flag = 0;
	PCERT_ENHKEY_USAGE pUsage = NULL;
	//check certificate OID for without encryption mode
	//need to set up the timeout and check if need without_enc mode
	if (!CertGetEnhancedKeyUsage (pCert, 0, NULL, &cbUsage)){
	    CertFreeCertificateContext(pCert);
	    return c->opt->option.without_enc?FALSE:TRUE;
	}
	pUsage = (PCERT_ENHKEY_USAGE)malloc(cbUsage);
	if (!CertGetEnhancedKeyUsage (pCert, 0, pUsage, &cbUsage))
	{
	    CertFreeCertificateContext(pCert);
	    free(pUsage);
	    return c->opt->option.without_enc?FALSE:TRUE;
	}
	for (i = 0; i < pUsage->cUsageIdentifier; i++){
	    if (strcmp(pUsage->rgpszUsageIdentifier[i], szOID_CP_HSMCA_COMPUTER_USER))
		continue;
	    flag = 1;
	}
	 if(0 == flag){
       
	    CertFreeCertificateContext(pCert);
	    free(pUsage);
	    if(c->opt->option.without_enc){
		s_log(LOG_ERR,"Client don't have szOID_CP_HSMCA_COMPUTER_USER, but try connect in \"without encryption\" mode");
		return FALSE;
	    }else 
		return TRUE;
	}else {
	    c->opt->timeout_idle = -1; 
	}
	free(pUsage);
    }
#endif
    CertFreeCertificateContext(pCert);
    s_log(LOG_DEBUG,"validate_connect finish");
    return TRUE;
}
//Implement verify remote cert 
static BOOL ssl_verify_by_level(CLI* c, PCCERT_CONTEXT pCert){
    HCERTSTORE hCertStore = NULL;
    PCCERT_CONTEXT pCertContext = NULL;
    CERT_CHAIN_PARA ChainPara;
    PCCERT_CHAIN_CONTEXT pChain=NULL;
    HTTPSPolicyCallbackData  polHttps;
    CERT_CHAIN_POLICY_PARA   PolicyPara;
    CERT_CHAIN_POLICY_STATUS PolicyStatus;
    LPSTR rgszUsages[] = { szOID_PKIX_KP_CLIENT_AUTH };
    DWORD cUsages = sizeof(rgszUsages) / sizeof(LPSTR);

    //if verify = 3 require presence of certificate in TrustedUsers
    if (c->opt->verify_use_only_my == 1){
#ifdef USE_WIN32
        hCertStore = CertOpenSystemStore(0, L"TrustedUsers");
#else
        hCertStore = CertOpenSystemStore(0, "TrustedUsers");
#endif
        if(!hCertStore){
            s_log(LOG_ERR,"Error 0x%x returned by CertOpenSystemStore \"TrustedUsers \" \n", 
            GetLastError());
            return SEC_E_NO_CREDENTIALS;
        }
        pCertContext = CertFindCertificateInStore(hCertStore, 
                                                  X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, 
                                                  0,
                                                  CERT_FIND_EXISTING,
                                                  pCert,
                                                  NULL);
        if(pCertContext == NULL)
        {
            CertCloseStore(hCertStore,0);
            s_log(LOG_ERR,"Error 0x%x returned by CertFindCertificateInStore \"TrustedUsers\" \n",
            GetLastError());
            return SEC_E_NO_CREDENTIALS;
        }
        CertFreeCertificateContext(pCertContext);
        CertCloseStore(hCertStore,0);
    }

    //Verify CertificateChain
    DisplayCertChain(pCert, TRUE);

    ChainPara.cbSize=sizeof(ChainPara);
    ChainPara.RequestedUsage.dwType=USAGE_MATCH_TYPE_AND;
    ChainPara.RequestedUsage.Usage.cUsageIdentifier = cUsages;
    ChainPara.RequestedUsage.Usage.rgpszUsageIdentifier = rgszUsages;

    if(!CertGetCertificateChain(
	  NULL,
	  pCert,
	  NULL,
	  NULL,
	  &ChainPara,
	  0,
	  NULL,
	  &pChain))
          {
              s_log(LOG_ERR,"Error 0x%x returned by CertGetCertificateChain\n", GetLastError());
              return SEC_E_NO_CREDENTIALS;
          }
    if(pChain->TrustStatus.dwErrorStatus){
        DisplayWinVerifyTrustError(pChain->TrustStatus.dwErrorStatus); 
        s_log(LOG_ERR,"Error 0x%x when validate certificate\n", pChain->TrustStatus.dwErrorStatus);
        CertFreeCertificateChain(pChain);
        return SEC_E_NO_CREDENTIALS;
    }

    s_log(LOG_NOTICE,"No error on CertGetCertificateChain");

    ZeroMemory(&polHttps, sizeof(HTTPSPolicyCallbackData));
    polHttps.cbStruct           = sizeof(HTTPSPolicyCallbackData);
    polHttps.dwAuthType         = AUTHTYPE_CLIENT;

    memset(&PolicyPara, 0, sizeof(PolicyPara));
    PolicyPara.cbSize            = sizeof(PolicyPara);
    PolicyPara.pvExtraPolicyPara = &polHttps;

    memset(&PolicyStatus, 0, sizeof(PolicyStatus));
    PolicyStatus.cbSize = sizeof(PolicyStatus);

    if(!CertVerifyCertificateChainPolicy(
                            CERT_CHAIN_POLICY_SSL,
                            pChain,
                            &PolicyPara,
                            &PolicyStatus))
    {
        s_log(LOG_ERR,"Error 0x%x returned by CertVerifyCertificateChainPolicy!", GetLastError());
        CertFreeCertificateChain(pChain);
        return SEC_E_NO_CREDENTIALS;
    }

    if(PolicyStatus.dwError){
        DisplayWinVerifyTrustError(PolicyStatus.dwError); 
        CertFreeCertificateChain(pChain);
        return SEC_E_NO_CREDENTIALS;
    }

    s_log(LOG_NOTICE,"No error on CertVerifyCertificateChainPolicy");
    CertFreeCertificateChain(pChain);
    return SEC_E_OK;
}
//Implement Check of certificate in handhsake phase
//On HSM_SERVER use pam to authenticate and use seteuid to transition to user from cert
static void verify_remote_cert(CLI* c){
#ifdef USE_PAM
struct pam_conv st_pam_conv;
int pam_status;
#endif
#ifdef HSM_SERVER
#ifdef SOLARIS
    void* uid;
#else
    const void* uid;
#endif
    uid_t user = 0;
    gid_t grp = 0;
#endif //HSM_SERVER
    
#ifdef USE_PAM
    st_pam_conv.conv = st_pamconv_func;
    st_pam_conv.appdata_ptr = NULL;    
    // use pam module to authentification       
    if((pam_status = pam_start("stunnel",NULL,&st_pam_conv, &c->pamh))!= PAM_SUCCESS){
          s_log(LOG_ERR,"pam_start failed err = %d", pam_status);
          longjmp(c->err,1);
    }
#endif //USE_PAM
   // get certificate  from schannel and put CN to PAM_USER
   // for HSM_SERVER check for extension
    if(!validate_connect(c)){
          s_log(LOG_NOTICE,"User not authorized for connect");
          c->ssl_wfd = &c->local_rfd;
          SSPI_shutdown(c);
          longjmp(c->err,1);
    }
#ifdef HSM_SERVER  
    if (geteuid())
        seteuid(0);
#endif
#ifdef USE_PAM
    s_log(LOG_DEBUG,"start pam_auth");
    if((pam_status = pam_authenticate(c->pamh,0)) != PAM_SUCCESS){
            s_log(LOG_ERR,"pam_authenticate failed err = %d", pam_status);
            pam_close_session(c->pamh,0);
            longjmp(c->err,1);
    }
    s_log(LOG_DEBUG,"pam_auth finish");
    if((pam_status = pam_open_session(c->pamh,0)) !=PAM_SUCCESS)
         s_log(LOG_ERR,"pam_open_session error = %d",pam_status);
#endif
    // we can get user from pam only on HSM_SERVER
#ifdef HSM_SERVER
    pam_get_item(c->pamh,PAM_USER,&uid);
    s_log(LOG_DEBUG,"uid = %s",(char *)uid);
    sscanf( uid, "%d.%d", &user, &grp );
    if ( !grp )
	grp = user;
//    user = atoi((const char *)uid);
    if(setegid(grp) != 0){ 
        s_log(LOG_ERR,"setgid error = %d", errno);
        longjmp(c->err,1);
    }
    if(seteuid(user) != 0){
        s_log(LOG_ERR,"seteuid error = %d", errno);
        longjmp(c->err,1);
    }
#endif //HSM_SERVER    
}
#endif //SSPI

/* End of client.c */
