/*
 *   stunnel       Universal SSL tunnel
 *   Copyright (c) 1998-2006 Michal Trojnara <Michal.Trojnara@mirt.net>
 *                 All Rights Reserved
 *
 *   Version:      4.18             (stunnel.c)
 *   Date:         2006.09.26
 *
 *   Author:       Michal Trojnara  <Michal.Trojnara@mirt.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   In addition, as a special exception, Michal Trojnara gives
 *   permission to link the code of this program with the OpenSSL
 *   library (or with modified versions of OpenSSL that use the same
 *   license as OpenSSL), and distribute linked combinations including
 *   the two.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than OpenSSL.  If you modify
 *   this file, you may extend this exception to your version of the
 *   file, but you are not obligated to do so.  If you do not wish to
 *   do so, delete this exception statement from your version.
 */

#include "common.h"
#include "prototypes.h"
#include <locale.h>
#ifdef IOS
#include <pthread.h>
#endif //IOS

    /* Prototypes */
static void daemon_loop(void);
static void accept_connection(LOCAL_OPTIONS *);
static void get_limits(void); /* setup global max_clients and max_fds */
#ifdef SSPI
#include "reader/support.h"
static HRESULT
CreateServCredentials(
    PCredHandle phCreds, //out
    LOCAL_OPTIONS *opt   //service options
    ); 

    #ifdef IOS
    extern DWORD CSP_GetLastError(void);
    #define GetLastError() CSP_GetLastError()
    #endif
#endif /* SSPI */

#if !defined (USE_WIN32) && !defined (__vms)
static SOCKET prepare_socket(const char *port_name);
static void drop_privileges(void);
static void daemonize(void);
static void create_pid(void);
static void delete_pid(void);
#endif

    /* Error/exceptions handling functions */
#ifndef USE_WIN32
static void signal_handler(int);
#endif

int volatile num_clients=0; /* Current number of clients */

    /* Functions */

#ifndef STUNNEL_MAIN
#ifdef IOS
#define STUNNEL_MAIN stunnel_main
#else //IOS
#define STUNNEL_MAIN main
#endif //IOS
#endif

#ifndef USE_WIN32
int STUNNEL_MAIN(int argc, char* argv[]) { /* execution begins here 8-) */

    main_initialize(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL);

    signal(SIGPIPE, SIG_IGN); /* avoid 'broken pipe' signal */
    if(signal(SIGTERM, SIG_IGN)!=SIG_IGN)
        signal(SIGTERM, signal_handler);
    if(signal(SIGQUIT, SIG_IGN)!=SIG_IGN)
        signal(SIGQUIT, signal_handler);
    if(signal(SIGINT, SIG_IGN)!=SIG_IGN)
        signal(SIGINT, signal_handler);
    if(signal(SIGHUP, SIG_IGN)!=SIG_IGN)
        signal(SIGHUP, signal_handler);
    /* signal(SIGSEGV, signal_handler); */

    main_execute();

    return 0; /* success */
}
#endif

void main_initialize(char *arg1, char *arg2) {
    setlocale(LC_ALL,"");
#ifdef HSM_SERVER
#ifdef UNIX
    support_set_impersonate(0,1,0);
#endif
#endif
#ifndef SSPI // do not need this on SSPI
    ssl_init(); /* initialize SSL library */
#endif //SSPI    
    sthreads_init(); /* initialize critical sections & SSL callbacks */
    parse_config(arg1, arg2);
    log_open();
    stunnel_info(0);
}

void main_execute(void) {
#ifdef SSPI
    if(!LoadSecurityLibrary())
    {
        s_log(LOG_ERR, "Error initializing the security library\n");
        exit(1);
    }
#ifdef USE_FORK
    {
        HCRYPTPROV hProv;
        DWORD param;
        DWORD cbData = 4;
        CryptAcquireContext(&hProv, NULL, NULL, 75, CRYPT_VERIFYCONTEXT);
        CryptGetProvParam(hProv, PP_SECURITY_LEVEL, (BYTE*)&param, &cbData, 0);
        CryptReleaseContext(hProv, 0);
        if(param == 1){
            s_log(LOG_ERR, "Default provider level is KC1. use stunnel_thread ONLY!");
            exit(1);
        }
    }
#endif //USE_FORK
#endif //SSPI

    /* check if started from inetd */
    if(local_options.next) { /* there are service sections -> daemon mode */
        daemon_loop();
    } else { /* inetd mode */
#if !defined (USE_WIN32) && !defined (__vms)&&!defined(USE_OS2)
        max_fds=FD_SETSIZE; /* just in case */
        drop_privileges();
#endif
        num_clients=1;
        client(alloc_client_session(&local_options, 0, 1));
    }
    log_close();
}

static void daemon_loop(void) {
    SOCKADDR_UNION addr;
    s_poll_set fds;
    LOCAL_OPTIONS *opt;
    get_limits();
    s_poll_zero(&fds);
#if !defined(USE_WIN32) && !defined(USE_OS2)
    s_poll_add(&fds, signal_pipe_init(), 1, 0);
#endif
    if(!local_options.next) {
        s_log(LOG_ERR, "No connections defined in config file");
        exit(1);
    }
    num_clients=0;

    /* bind local ports */
    for(opt=local_options.next; opt; opt=opt->next) {
        if(!opt->option.accept) /* no need to bind this service */
            continue;
#ifndef USE_WIN32
	if(opt->local_sockaddr){
            s_log(LOG_DEBUG,"accept addres =%s",opt->local_sockaddr);
            opt->fd = prepare_socket(opt->local_sockaddr);
            if(opt->fd == INVALID_SOCKET){
            s_log(LOG_ERR,"Prepare unix socket error");
            exit(1);
        }
        } else{ 
#endif //USE_WIN32
            memcpy(&addr, &opt->local_addr.addr[0], sizeof(SOCKADDR_UNION));
            if((opt->fd=socket(addr.sa.sa_family, SOCK_STREAM, 0))<0) {
                sockerror("local socket");
                exit(1);
            }
            if(alloc_fd(opt->fd))
                exit(1);
            if(set_socket_options(opt->fd, 0)<0)
                exit(1);
            s_ntop(opt->local_address, &addr);
            if(bind(opt->fd, &addr.sa, addr_len(addr))) {
                s_log(LOG_ERR, "Error binding %s to %s",
                    opt->servname, opt->local_address);
                sockerror("bind");
                exit(1);
            }
            s_log(LOG_DEBUG, "%s bound to %s", opt->servname, opt->local_address);
            if(st_listen(opt->fd, SOMAXCONN)) {
                sockerror("listen");
                exit(1);
            }
#ifdef FD_CLOEXEC
        fcntl(opt->fd, F_SETFD, FD_CLOEXEC); /* close socket in child execvp */
#endif
#ifndef USE_WIN32
	}
#endif
        s_poll_add(&fds, opt->fd, 1, 0);
    }

#if !defined (USE_WIN32) && !defined (__vms) && !defined(USE_OS2)
    if(!(options.option.foreground))
        daemonize();
    drop_privileges();
    create_pid();
#endif /* !defined USE_WIN32 && !defined (__vms) */
    /* create exec+connect services */
    for(opt=local_options.next; opt; opt=opt->next) {
#ifdef SSPI
	if(!opt->option.client){
#ifdef HSM_SERVER
    if(!(opt->setegid_group && opt->seteuid_user)){ //all tls keys under 999 user on HSM
	setegid(999);seteuid(999);
    }
#endif
	    if(FAILED(CreateServCredentials(&opt->hServerCreds,opt)))
	    {
		s_log(LOG_ERR,"Error creating credentials\n");
		exit(1);
	    }
#ifdef HSM_SERVER
    if(!(opt->setegid_group && opt->seteuid_user)){
	seteuid(0);setegid(0);
    }
#endif
	    s_log(LOG_DEBUG,"CreateServCreadentials finish");
	}
#endif //SSPI	
        if(opt->option.accept) /* skip ordinary (accepting) services */
            continue;
        enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
        ++num_clients;
        leave_critical_section(CRIT_CLIENTS);
        create_client(-1, -1, alloc_client_session(opt, -1, -1), client);
    }
    //Notify Sc so we are running
#ifdef USE_WIN32
    ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 0);
#endif
    while(1) {
        if(s_poll_wait(&fds, -1, -1)<0) { /* non-critical error */
            log_error(LOG_INFO, get_last_socket_error(),
                "daemon_loop: s_poll_wait");
            sleep(1); /* to avoid log trashing */
        } else {
            for(opt=local_options.next; opt; opt=opt->next)
                if(s_poll_canread(&fds, opt->fd))
                    accept_connection(opt);

        }
    }
    s_log(LOG_ERR, "INTERNAL ERROR: End of infinite loop 8-)");
}

static void accept_connection(LOCAL_OPTIONS *opt) {
    SOCKADDR_UNION addr;
    char from_address[IPLEN];
    SOCKET s;
    socklen_t addrlen;

    addrlen=sizeof(SOCKADDR_UNION);
    while((s=st_accept(opt->fd, &addr.sa, &addrlen))<0) {
        switch(get_last_socket_error()) {
            case EINTR:
                break; /* retry */
            case EMFILE:
#ifdef ENFILE
            case ENFILE:
#endif
#ifdef ENOBUFS
            case ENOBUFS:
#endif
            case ENOMEM:
                sleep(1); /* temporarily out of resources - short delay */
            default:
                sockerror("accept");
                return; /* error */
        }
    }
    s_ntop(from_address, &addr);
    s_log(LOG_DEBUG, "%s accepted FD=%d from %s",
        opt->servname, s, from_address);
    if(max_clients && num_clients>=max_clients) {
        s_log(LOG_WARNING, "Connection rejected: too many clients (>=%d)",
            max_clients);
        closesocket(s);
        return;
    }
#ifdef FD_CLOEXEC
    fcntl(s, F_SETFD, FD_CLOEXEC); /* close socket in child execvp */
#endif
    if(create_client(opt->fd, s, alloc_client_session(opt, s, s), client)) {
        s_log(LOG_ERR, "Connection rejected: create_client failed");
        closesocket(s);
        return;
    }
    enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
    ++num_clients;
    leave_critical_section(CRIT_CLIENTS);
}

static void get_limits(void) {
#ifdef USE_WIN32
    s_log(LOG_NOTICE, "No limit detected for the number of clients");
#else
#if defined(USE_OS2) && defined(__INNOTEK_LIBC__)
    /* OS/2 with the Innotek LIBC does not share the same handles between files
     and socket connections.
     */
    max_clients=(FD_SETSIZE-6)/2;
    max_fds=FD_SETSIZE;
#else
    max_fds=0; /* unlimited */

#if defined HAVE_SYSCONF
    max_fds=sysconf(_SC_OPEN_MAX);
    if(max_fds<0)
        ioerror("sysconf");
#elif defined HAVE_GETRLIMIT
    struct rlimit rlim;
    if(getrlimit(RLIMIT_NOFILE, &rlim)<0)
        ioerror("getrlimit");
    else
        max_fds=rlim.rlim_cur;
    if(max_fds==RLIM_INFINITY)
        max_fds=0; /* RLIM_INFINITY should be equal to zero, anyway */
#endif
    s_log(LOG_INFO, "file ulimit = %d%s (can be changed with 'ulimit -n')",
        max_fds, max_fds ? "" : " (unlimited)");
#ifdef USE_POLL
    s_log(LOG_INFO, "poll() used - no FD_SETSIZE limit for file descriptors");
#else
    s_log(LOG_INFO,
        "FD_SETSIZE = %d (some systems allow to increase this value)",
        FD_SETSIZE);
    if(!max_fds || max_fds>FD_SETSIZE)
        max_fds=FD_SETSIZE;
#endif
    if(max_fds && max_fds<16) /* stunnel needs at least 16 file desriptors */
        max_fds=16;
    if(max_fds) {
        if(max_clients>(max_fds>=256 ? max_fds*125/256 : (max_fds-6)/2))
            max_clients=max_fds>=256 ? max_fds*125/256 : (max_fds-6)/2;
        s_log(LOG_NOTICE, "%d clients allowed", max_clients);
    } else {
        s_log(LOG_NOTICE, "No limit detected for the number of clients");
    }
#endif
#endif
}

#if !defined (USE_WIN32) && !defined (__vms)
    /* chroot and set process user and group(s) id */
static void drop_privileges(void) {
    int uid=0, gid=0;
    struct group *gr;
#ifdef HAVE_SETGROUPS
    gid_t gr_list[1];
#endif
    struct passwd *pw;

    /* Get the integer values */
    if(options.setgid_group) {
        gr=getgrnam(options.setgid_group);
        if(gr)
            gid=gr->gr_gid;
        else if(atoi(options.setgid_group)) /* numerical? */
            gid=atoi(options.setgid_group);
        else {
            s_log(LOG_ERR, "Failed to get GID for group %s",
                options.setgid_group);
            exit(1);
        }
    }
    if(options.setuid_user) {
        pw=getpwnam(options.setuid_user);
        if(pw)
            uid=pw->pw_uid;
        else if(atoi(options.setuid_user)) /* numerical? */
            uid=atoi(options.setuid_user);
        else {
            s_log(LOG_ERR, "Failed to get UID for user %s",
                options.setuid_user);
            exit(1);
        }
    }

#ifdef HAVE_CHROOT
    /* chroot */
    if(options.chroot_dir) {
        if(chroot(options.chroot_dir)) {
            sockerror("chroot");
            exit(1);
        }
        if(chdir("/")) {
            sockerror("chdir");
            exit(1);
        }
    }
#endif /* HAVE_CHROOT */

    /* Set uid and gid */
    if(gid) {
        if(setgid(gid)) {
            sockerror("setgid");
            exit(1);
        }
#ifdef HAVE_SETGROUPS
        gr_list[0]=gid;
        if(setgroups(1, gr_list)) {
            sockerror("setgroups");
            exit(1);
        }
#endif
    }
    if(uid) {
        if(setuid(uid)) {
            sockerror("setuid");
            exit(1);
        }
    }
}

static void daemonize(void) { /* go to background */
#if defined(HAVE_DAEMON) && !defined(__BEOS__)
    if(daemon(0,0)==-1) {
        ioerror("daemon");
        exit(1);
    }
#else
    chdir("/");
    switch(fork()) {
    case -1:    /* fork failed */
        ioerror("fork");
        exit(1);
    case 0:     /* child */
        break;
    default:    /* parent */
        exit(0);
    }
    close(0);
    close(1);
    close(2);
#endif
#ifdef HAVE_SETSID
    setsid(); /* Ignore the error */
#endif
}

static void create_pid(void) {
    int pf;
    char pid[STRLEN];

    if(!options.pidfile) {
        s_log(LOG_DEBUG, "No pid file being created");
        return;
    }
    if(options.pidfile[0]!='/') {
        s_log(LOG_ERR, "Pid file (%s) must be full path name", options.pidfile);
        /* Why?  Because we don't want to confuse by
           allowing '.', which would be '/' after
           daemonizing) */
        exit(1);
    }
    options.dpid=(unsigned long)getpid();

    /* silently remove old pid file */
    unlink(options.pidfile);
    if((pf=open(options.pidfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL,0644))==-1) {
        s_log(LOG_ERR, "Cannot create pid file %s", options.pidfile);
        ioerror("create");
        exit(1);
    }
    sprintf(pid, "%lu\n", options.dpid);
    write(pf, pid, strlen(pid));
    close(pf);
    s_log(LOG_DEBUG, "Created pid file %s", options.pidfile);
    atexit(delete_pid);
}

static void delete_pid(void) {
    s_log(LOG_DEBUG, "removing pid file %s", options.pidfile);
    if((unsigned long)getpid()!=options.dpid)
        return; /* current process is not main daemon process */
    if(unlink(options.pidfile)<0)
        ioerror(options.pidfile); /* not critical */
}

static void signal_handler(int sig) { /* signal handler */
    s_log(sig==SIGTERM ? LOG_NOTICE : LOG_ERR,
        "Received signal %d; terminating", sig);
#ifdef IOS
    int retval = 3;
    pthread_exit(&retval);
#else //IOS
    exit(3);
#endif //IOS
}
#endif /* !defined USE_WIN32 */

void stunnel_info(int raw) {
    char line[STRLEN];

#ifndef SSPI
    sprintf(line, "stunnel " VERSION " on " HOST " with %s",
        SSLeay_version(SSLEAY_VERSION));
#else
    sprintf(line, "stunnel " VERSION " on " HOST 
        );
#endif //SSPI
    if(raw)
        log_raw("%s", line);
    else
        s_log(LOG_NOTICE, "%s", line);

    safecopy(line, "Threading:");
#ifdef USE_UCONTEXT
    safeconcat(line, "UCONTEXT");
#endif
#ifdef USE_PTHREAD
    safeconcat(line, "PTHREAD");
#endif
#ifdef USE_WIN32
    safeconcat(line, "WIN32");
#endif
#ifdef USE_FORK
    safeconcat(line, "FORK");
#endif

#ifdef HAVE_OSSL_ENGINE_H
    safeconcat(line, " SSL:ENGINE");
#endif

    safeconcat(line, " Sockets:");
#ifdef USE_POLL
    safeconcat(line, "POLL");
#else /* defined(USE_POLL) */
    safeconcat(line, "SELECT");
#endif /* defined(USE_POLL) */
//#TODO:       !defined(HAVE_GETADDRINFO)
#if defined(USE_WIN32) && !defined(_WIN32_WCE) && !defined(HAVE_GETADDRINFO)
    if(s_getaddrinfo)
        safeconcat(line, ",IPv6");
    else
        safeconcat(line, ",IPv4");
#else /* defined(USE_WIN32) */
#if defined(USE_IPv6)
    safeconcat(line, ",IPv6");
#else /* defined(USE_IPv6) */
    safeconcat(line, ",IPv4");
#endif /* defined(USE_IPv6) */
#endif /* defined(USE_WIN32) */

#ifdef USE_LIBWRAP
    safeconcat(line, " Auth:LIBWRAP");
#endif

    if(raw)
        log_raw("%s", line);
    else
        s_log(LOG_NOTICE, "%s", line);
}
#ifdef SSPI
//-------------------------------------------------------------
static
HRESULT
CreateServCredentials(
    PCredHandle phCreds,            // out
    LOCAL_OPTIONS *opt)   //service options
{
    HCERTSTORE      hMyCertStore = NULL;
    HCERTSTORE      hSysCertStore = NULL;

    DWORD   dwProtocol      = 0;

    SCHANNEL_CRED   SchannelCred;
    TimeStamp       tsExpiry;
    SECURITY_STATUS Status;
    PCCERT_CONTEXT  pCertContext = NULL;
    PCCERT_CONTEXT  pCert = NULL;

    get_cert_from_file(opt->cert, &pCert);
    if(!pCert){
        s_log(LOG_ERR,"CertCreateCertificateContext failed: %ud\n", GetLastError( ) );
        return SEC_E_NO_CREDENTIALS;
    }
#ifdef HSM_SERVER
    if(opt->setegid_group && opt->seteuid_user){
	setegid(opt->setegid_group); 
	seteuid(opt->seteuid_user); 
    }
#endif
    //Try System store then current user
#ifndef HSM_SERVER
    hSysCertStore = CertOpenStore(CERT_STORE_PROV_SYSTEM,PKCS_7_ASN_ENCODING|X509_ASN_ENCODING,
		    0,CERT_SYSTEM_STORE_LOCAL_MACHINE,L"MY");
#endif
#ifdef UNICODE
    hMyCertStore = CertOpenSystemStore(0, L"MY");
#else
    hMyCertStore = CertOpenSystemStore(0, "MY");
#endif

    if(!hMyCertStore && !hSysCertStore){
        s_log(LOG_ERR,"Error 0x%x returned by CertOpenSystemStore(My) and CertOpenStore(System)\n", 
        GetLastError());
        CertFreeCertificateContext(pCert);
#ifdef HSM_SERVER
	seteuid (0); setegid (0);
#endif
        return SEC_E_NO_CREDENTIALS;
    }
#ifndef HSM_SERVER
    pCertContext = CertFindCertificateInStore(hSysCertStore, 
                                              X509_ASN_ENCODING|PKCS_7_ASN_ENCODING, 
                                              0,
                                              CERT_FIND_EXISTING,
                                              pCert,
                                              NULL);
#endif
    //There is no required certificate in System store => search current user
    if(pCertContext == NULL){
	s_log(LOG_NOTICE,"CertFindCertificateInStore not find certificate in LOCAL_MACHINE store. Looking at CURRENT_USER");
	pCertContext = CertFindCertificateInStore(hMyCertStore, 
                                              X509_ASN_ENCODING|PKCS_7_ASN_ENCODING, 
                                              0,
                                              CERT_FIND_EXISTING,
                                              pCert,
                                              NULL);
    }
    if(pCertContext == NULL)
    {
        s_log(LOG_ERR,"Error 0x%x returned by CertFindCertificateInStore",
        GetLastError());
        CertFreeCertificateContext(pCert);
	if(hSysCertStore) CertCloseStore(hSysCertStore,0);
	if(hMyCertStore) CertCloseStore(hMyCertStore,0);
#ifdef HSM_SERVER
	seteuid (0); setegid (0);
#endif
        return SEC_E_NO_CREDENTIALS;
    }
    CertFreeCertificateContext(pCert);
    if(opt->pincode){
        HCRYPTPROV hProv = 0;
        DWORD keytype = 0;
        BOOL should_release_ctx = FALSE;
        s_log(LOG_NOTICE,"pincode option is present. Call CryptSetProvParam");
        Status = CryptAcquireCertificatePrivateKey(pCertContext,0,NULL,&hProv,&keytype,&should_release_ctx);
        if(!Status){
            s_log(LOG_ERR,"CryptAcquireCertificatePrivateKey failed. Error = 0x%x", GetLastError());
            return SEC_E_NO_CREDENTIALS;
        }
        Status = CryptSetProvParam(hProv,PP_KEYEXCHANGE_PIN,(BYTE*)opt->pincode,0);
        if(!Status){
            BYTE buff[255];
	    DWORD len = sizeof(buff);
	    s_log(LOG_ERR,"CryptSetProvParam PP_KEYEXCHANGE_PIN failed. Error = 0x%x, pincode=%s", GetLastError(),opt->pincode);
	    CryptGetProvParam(hProv,PP_CONTAINER,buff,&len,0);
	    s_log(LOG_ERR,"Container Name = %s",(char*)buff);
	    return SEC_E_NO_CREDENTIALS;
	}
	s_log(LOG_NOTICE,"Call CryptSetProvParam change pin OK");
    }
    ZeroMemory(&SchannelCred, sizeof(SchannelCred));
    SchannelCred.dwVersion = SCHANNEL_CRED_VERSION;
    SchannelCred.cCreds = 1;
    SchannelCred.paCred = &pCertContext;
    SchannelCred.grbitEnabledProtocols = dwProtocol;

    Status = g_pSSPI->AcquireCredentialsHandle(
                        NULL,                   
                        UNISP_NAME,           
                        SECPKG_CRED_INBOUND,    
                        NULL,                   
                        &SchannelCred,          
                        NULL,                   
                        NULL,                   
                        phCreds,                
                        &tsExpiry);             
#ifdef HSM_SERVER
    if(opt->setegid_group && opt->seteuid_user){
	seteuid (0);
	setegid (0);
    }
#endif
    if(Status != SEC_E_OK)
    {
        s_log(LOG_ERR,"**** Error 0x%x returned by AcquireCredentialsHandle\n", Status);
        CertFreeCertificateContext(pCertContext);
	if(hSysCertStore) CertCloseStore(hSysCertStore,0);
	if(hMyCertStore) CertCloseStore(hMyCertStore,0);
        return Status;
    }
    if(pCertContext)
    {
        CertFreeCertificateContext(pCertContext);
    }
    if(hSysCertStore) CertCloseStore(hSysCertStore,0);
    if(hMyCertStore) CertCloseStore(hMyCertStore,0);
    return SEC_E_OK;
} //CreateServCredentials
#ifdef _WIN32
VOID ServiceStart(DWORD dwArgc, LPTSTR *lpszArgv){
    static struct WSAData wsa_state;
    if(WSAStartup(MAKEWORD(1, 1), &wsa_state))
        return;
#ifdef UNICODE
    {
	
    //convert to char
    char arg1[255];
    char arg2[255];
    memset(arg1,0,255);
    memset(arg2,0,255);
    if(dwArgc>1){
	safe_wcsrtombs(arg1,lpszArgv[1],wcslen(lpszArgv[1]));
    }
    if(dwArgc>2){
	safe_wcsrtombs(arg2,lpszArgv[2],wcslen(lpszArgv[2]));
    }
  //  log_raw(" Service start argument1 %s, argument2 %s", arg1,arg2);
    main_initialize(dwArgc>1 ? arg1 : NULL, dwArgc>2 ? arg2 : NULL);
    }
#else
    main_initialize(dwArgc>1 ? lpszArgv[1] : NULL, dwArgc>2 ? lpszArgv[2] : NULL);
#endif
    main_execute();
};
VOID ServiceStop(){
    ReportStatusToSCMgr(SERVICE_STOPPED, NO_ERROR,0);
};
#endif// _WIN32
#endif //SSPI
#ifdef UNIX
static SOCKET prepare_socket(const char *port_name)
{
    SOCKET s = INVALID_SOCKET;
    int rc;
    struct sockaddr_un serv_addr;

    s = socket(AF_UNIX, SOCK_STREAM, 0);
    if( s == INVALID_SOCKET ) {
	s_log(LOG_ERR,": socket() fail: %s", strerror(errno));
	return -1;
    }

    memset( &serv_addr, 0, sizeof(serv_addr));
    serv_addr.sun_family = AF_UNIX;
    strcpy (serv_addr.sun_path, port_name);
    if (remove (port_name) && errno != ENOENT) {
	s_log(LOG_ERR,": remove() fail: %s", strerror(errno));
	return -1;
    }
    s_log(LOG_DEBUG,"socket for bind %s",serv_addr.sun_path);
    rc = bind(s, (struct sockaddr *) &serv_addr, (int)sizeof(serv_addr));
    if( rc == -1 ) {
	s_log(LOG_ERR, ": bind() fail: %s", strerror(errno));
	return -1;
    }
    if (chmod (port_name, 0777)) {
	s_log(LOG_ERR, ": chmod(%s,0777) fail: %s", port_name, strerror(errno));
	return -1;
    }
    if(st_listen(s, 10) == -1) {
	s_log(LOG_ERR,": listen() fail: %s", strerror(errno));
	return -1;
    }

    return s;
}//prepare_socket
#endif //UNIX

/* End of stunnel.c */
